#    Copyright (c) 2014 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import json
import multiprocessing
import socket
import time

import eventlet
from oslo_config import cfg
from oslo_log import log as logging
from oslo_service import service
import psutil
from sqlalchemy import desc

from murano.api import v1
from murano.api.v1.deployments import set_dep_state
from murano.api.v1 import request_statistics
from murano.db import models
from murano.db.services import environments as envs
from murano.db.services import stats as db_stats
from murano.db import session as db_session
from murano.engine.system import status_reporter

CONF = cfg.CONF

CONF_STATS = CONF.stats
LOG = logging.getLogger(__name__)


class StatsCollectingService(service.Service):
    def __init__(self):
        super(StatsCollectingService, self).__init__()
        request_statistics.init_stats()
        self._hostname = socket.gethostname()
        self._stats_db = db_stats.Statistics()
        self._prev_time = time.time()
        self._notifier = status_reporter.Notification()

    def start(self):
        super(StatsCollectingService, self).start()
        self.tg.add_thread(self._collect_stats_loop)
        self.tg.add_thread(self._report_env_stats_loop)

    def stop(self):
        super(StatsCollectingService, self).stop()

    def _collect_stats_loop(self):
        period = CONF_STATS.period * 60
        while True:
            self.update_stats()
            eventlet.sleep(period)

    def _report_env_stats_loop(self):
        env_audit_period = CONF_STATS.env_audit_period * 60
        while True:
            self.report_env_stats()
            eventlet.sleep(env_audit_period)

    def update_stats(self):
        LOG.debug("Updating statistic information.")
        LOG.debug("Stats object: {stats}".format(stats=v1.stats))
        LOG.debug("Stats: (Requests: {amount}  Errors: {error}  "
                  "Ave.Res.Time {time:2.4f}\n Per tenant: {req_count})".format(
                      amount=v1.stats.request_count,
                      error=v1.stats.error_count,
                      time=v1.stats.average_time,
                      req_count=v1.stats.requests_per_tenant))
        try:
            stats = self._stats_db.get_stats_by_host(self._hostname)
            if stats is None:
                self._stats_db.create(self._hostname,
                                      v1.stats.request_count,
                                      v1.stats.error_count,
                                      v1.stats.average_time,
                                      v1.stats.requests_per_tenant,
                                      multiprocessing.cpu_count(),
                                      psutil.cpu_percent())
                return

            now = time.time()
            t_delta = now - self._prev_time
            requests_per_second = (v1.stats.request_count -
                                   stats.request_count) / t_delta
            errors_per_second = (v1.stats.error_count -
                                 stats.error_count) / t_delta
            self._prev_time = now
            stats.request_count = v1.stats.request_count
            stats.error_count = v1.stats.error_count
            stats.average_response_time = v1.stats.average_time
            stats.requests_per_tenant = json.dumps(v1.stats.
                                                   requests_per_tenant)
            stats.requests_per_second = requests_per_second
            stats.errors_per_second = errors_per_second
            stats.cpu_percent = psutil.cpu_percent()
            self._stats_db.update(self._hostname, stats)
        except Exception as e:
            LOG.exception("Failed to get statistics object from a "
                          "database. {error_code}".format(error_code=e))

    def report_env_stats(self):
        LOG.debug("Reporting env stats")
        try:
            environments = envs.EnvironmentServices.get_environments_by({})

            for env in environments:
                deployments = get_env_deployments(env.id)
                success_deployments = [d for d in deployments
                                       if d['state'] == "success"]
                if success_deployments:
                    self._notifier.report('environment.exists', env.to_dict())
        except Exception:
            LOG.exception("Failed to report existing envs")


def get_env_deployments(environment_id):
    unit = db_session.get_session()
    query = unit.query(models.Task).filter_by(
        environment_id=environment_id).order_by(desc(models.Task.created))
    result = query.all()
    deployments = [
        set_dep_state(deployment, unit).to_dict() for deployment in result]
    return deployments



# Copyright (c) 2016 AT&T Corp
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from webob import exc

from murano.services import states
import murano.tests.unit.base as test_base
from murano.tests.unit import utils as test_utils
from murano import utils


class TestUtils(test_base.MuranoTestCase):

    @mock.patch('murano.utils.db_session')
    def test_check_env(self, mock_db_session):
        """Test check env."""
        mock_request = mock.MagicMock(context=test_utils.dummy_context())
        mock_env = mock.MagicMock(environment_id='test_env_id',
                                  tenant_id=mock_request.context.tenant)
        mock_db_session.get_session().query().get.return_value = mock_env

        env = utils.check_env(mock_request, mock_env.environment_id)
        self.assertEqual(mock_env, env)

    @mock.patch('murano.utils.db_session')
    def test_check_env_with_null_environment_id(self, mock_db_session):
        """Test check env with null environment id throws exception."""
        mock_request = mock.MagicMock(context=test_utils.dummy_context())
        mock_db_session.get_session().query().get.return_value = None

        test_env_id = 'test_env_id'
        expected_error_message = 'Environment with id {env_id} not found'\
                                 .format(env_id=test_env_id)

        with self.assertRaisesRegex(exc.HTTPNotFound,
                                    expected_error_message):
            utils.check_env(mock_request, test_env_id)

    @mock.patch('murano.utils.db_session')
    def test_check_env_with_mismatching_tenant_id(self, mock_db_session):
        """Test check env without matching tenant ids throws exception."""
        mock_request = mock.MagicMock(context=test_utils.dummy_context())
        mock_env = mock.MagicMock(environment_id='test_env_id',
                                  tenant_id='another_test_tenant_id')
        mock_db_session.get_session().query().get.return_value = mock_env

        expected_error_message = 'User is not authorized to access these '\
                                 'tenant resources'

        with self.assertRaisesRegex(exc.HTTPForbidden,
                                    expected_error_message):
            utils.check_env(mock_request, mock_env.environment_id)

    def test_check_session_with_null_session(self):
        """Test check session with null session throws exception."""
        expected_error_message = 'Session <SessionId {id}> is not found'\
                                 .format(id=None)
        with self.assertRaisesRegex(exc.HTTPNotFound,
                                    expected_error_message):
            utils.check_session(None, None, None, None)

    @mock.patch('murano.utils.check_env')
    def test_check_session_with_mismatching_environment_id(self, _):
        """Test check session without matching env ids throws exception."""
        mock_session = mock.MagicMock(session_id='session_id',
                                      environment_id='environment_id')
        environment_id = 'another_environment_id'
        expected_error_msg = 'Session <SessionId {session_id}> is not tied '\
                             'with Environment <EnvId {environment_id}>'\
                             .format(session_id=mock_session.session_id,
                                     environment_id=environment_id)
        with self.assertRaisesRegex(exc.HTTPBadRequest,
                                    expected_error_msg):
            utils.check_session(None, environment_id, mock_session,
                                mock_session.session_id)

    def test_verify_session_with_invalid_request_header(self):
        """Test session with invalid request header throws exception."""
        dummy_context = test_utils.dummy_context()
        if dummy_context.session:
            dummy_context.session = None
        mock_request = mock.MagicMock(context=dummy_context)
        expected_error_message = 'X-Configuration-Session header which '\
                                 'indicates to the session is missed'
        with self.assertRaisesRegex(exc.HTTPBadRequest,
                                    expected_error_message):
            self._test_verify_session(mock_request)

    @mock.patch('murano.utils.db_session')
    def test_verify_session_with_null_session(self, mock_db_session):
        """Test null session throws expected exception."""
        mock_request = mock.MagicMock(context=test_utils.dummy_context())
        mock_request.context.session = mock.MagicMock(
            return_value='test_sess_id')
        mock_db_session.get_session().query().get.return_value = None
        expected_error_message = 'Session <SessionId {0}> is not found'\
                                 .format(mock_request.context.session)
        with self.assertRaisesRegex(exc.HTTPNotFound,
                                    expected_error_message):
            self._test_verify_session(mock_request)

    @mock.patch('murano.utils.db_session')
    def test_verify_env_template_with_invalid_tenant(self, mock_db_session):
        """Test session validation failure throws expected exception."""
        mock_request = mock.MagicMock(context=test_utils.dummy_context())
        mock_request.context.tenant = mock.MagicMock(
            return_value='test_tenant_id')
        mock_env_template = mock.MagicMock(tenant_id='another_test_tenant_id')
        mock_db_session.get_session().query().get.return_value =\
            mock_env_template
        expected_error_message = 'User is not authorized to access this'\
                                 ' tenant resources'
        with self.assertRaisesRegex(exc.HTTPForbidden,
                                    expected_error_message):
            self._test_verify_env_template(mock_request, None)

    @mock.patch('murano.utils.db_session')
    def test_verify_env_template_with_null_template(self, mock_db_session):
        """Test null env template throws expected exception."""
        mock_db_session.get_session().query().get.return_value = None
        expected_error_message = 'Environment Template with id {id} not found'\
                                 .format(id='test_env_template_id')
        with self.assertRaisesRegex(exc.HTTPNotFound,
                                    expected_error_message):
            self._test_verify_env_template(None, 'test_env_template_id')

    @utils.verify_env_template
    def _test_verify_env_template(self, request, env_template_id):
        """Helper function for testing above decorator."""
        pass

    @mock.patch('murano.utils.sessions.SessionServices.validate')
    @mock.patch('murano.utils.db_session')
    def test_verify_session_with_invalid_session(self, mock_db_session,
                                                 mock_validate):
        """Test session validation failure throws expected exception."""
        mock_validate.return_value = False
        mock_request = mock.MagicMock(context=test_utils.dummy_context())
        mock_request.context.session = mock.MagicMock(
            return_value='test_sess_id')
        mock_db_session.get_session().query().get.return_value =\
            mock.MagicMock()
        expected_error_message = 'Session <SessionId {0}> is invalid: '\
                                 'environment has been updated or '\
                                 'updating right now with other session'\
                                 .format(mock_request.context.session)
        with self.assertRaisesRegex(exc.HTTPForbidden,
                                    expected_error_message):
            self._test_verify_session(mock_request)

    @mock.patch('murano.utils.sessions.SessionServices.validate')
    @mock.patch('murano.utils.db_session')
    def test_verify_session_in_deploying_state(self,
                                               mock_db_session,
                                               mock_validate):
        """Test deploying session throws expected exception."""
        mock_validate.return_value = True
        mock_request = mock.MagicMock(context=test_utils.dummy_context())
        mock_request.context.session = mock.MagicMock(
            return_value='test_sess_id')
        mock_db_session.get_session().query().get.return_value =\
            mock.MagicMock(state=states.SessionState.DEPLOYING)
        expected_error_message = 'Session <SessionId {0}> is already in '\
                                 'deployment state'\
                                 .format(mock_request.context.session)
        with self.assertRaisesRegex(exc.HTTPForbidden,
                                    expected_error_message):
            self._test_verify_session(mock_request)

    @utils.verify_session
    def _test_verify_session(self, request):
        """Helper function for testing above decorator."""
        pass



# Copyright 2016 AT&T Corp
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

import webob

from murano.api.middleware import version_negotiation
from murano.api import versions
from murano.tests.unit import base


class MiddlewareVersionNegotiationTest(base.MuranoTestCase):

    @mock.patch.object(version_negotiation, 'LOG')
    def test_middleware_version_negotiation_default(self, mock_log):
        middleware_vn = version_negotiation.VersionNegotiationFilter(None)
        request = webob.Request.blank('/environments')

        result = middleware_vn.process_request(request)

        self.assertIsInstance(result, versions.Controller)
        mock_log.warning.assert_called_once_with(
            "Unknown version. Returning version choices.")

    @mock.patch.object(version_negotiation, 'LOG')
    def test_process_request(self, mock_log):
        """Test process_request using different valid paths."""
        middleware_vn = version_negotiation.VersionNegotiationFilter(None)

        for path in ('v1', '/v1', '///v1'):
            request = webob.Request.blank(path)
            request.method = 'GET'
            request.environ = {}

            result = middleware_vn.process_request(request)

            self.assertIsNone(result)
            self.assertIn('api.version', request.environ)
            self.assertEqual(1, request.environ['api.version'])
            self.assertEqual('/v1', request.path_info)
            mock_log.debug.assert_any_call(
                "Matched version: v{version}".format(version=1))
            mock_log.debug.assert_any_call(
                'new path {path}'.format(path='/v1'))

        request = webob.Request.blank('/v1/')
        request.method = 'GET'
        request.environ = {}

        result = middleware_vn.process_request(request)

        self.assertIsNone(result)
        self.assertIn('api.version', request.environ)
        self.assertEqual(1, request.environ['api.version'])
        self.assertEqual('/v1/', request.path_info)
        mock_log.debug.assert_any_call(
            "Matched version: v{version}".format(version=1))
        mock_log.debug.assert_any_call(
            'new path {path}'.format(path='/v1/'))

    @mock.patch.object(version_negotiation, 'LOG')
    def test_process_request_without_path(self, mock_log):
        middleware_vn = version_negotiation.VersionNegotiationFilter(None)

        request = webob.Request.blank('')
        request.method = 'GET'
        request.environ = {}

        result = middleware_vn.process_request(request)

        self.assertIsInstance(result, versions.Controller)
        mock_log.warning.assert_called_once_with(
            "Unknown version. Returning version choices.")

    def test_factory(self):
        app = version_negotiation.VersionNegotiationFilter.factory(None)
        self.assertIsNotNone(app)
        self.assertEqual(
            version_negotiation.VersionNegotiationFilter, type(app(None)))



#    Copyright (c) 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from yaql.language import specs
from yaql.language import yaqltypes

from murano.dsl import dsl
from murano.dsl import helpers


@dsl.name('io.murano.system.GC')
class GarbageCollector(object):
    @staticmethod
    @specs.parameter('publisher', dsl.MuranoObjectParameter(decorate=False))
    @specs.parameter('subscriber', dsl.MuranoObjectParameter(decorate=False))
    @specs.parameter('handler', yaqltypes.String(nullable=True))
    def subscribe_destruction(publisher, subscriber, handler=None):
        publisher_this = publisher.real_this
        subscriber_this = subscriber.real_this

        if handler:
            subscriber.type.find_single_method(handler)

        dependency = GarbageCollector._find_dependency(
            publisher_this, subscriber_this, handler)

        if not dependency:
            dependency = {'subscriber': helpers.weak_ref(subscriber_this),
                          'handler': handler}
            publisher_this.destruction_dependencies.append(dependency)

    @staticmethod
    @specs.parameter('publisher', dsl.MuranoObjectParameter(decorate=False))
    @specs.parameter('subscriber', dsl.MuranoObjectParameter(decorate=False))
    @specs.parameter('handler', yaqltypes.String(nullable=True))
    def unsubscribe_destruction(publisher, subscriber, handler=None):
        publisher_this = publisher.real_this
        subscriber_this = subscriber.real_this

        if handler:
            subscriber.type.find_single_method(handler)

        dds = publisher_this.destruction_dependencies
        dependency = GarbageCollector._find_dependency(
            publisher_this, subscriber_this, handler)

        if dependency:
            dds.remove(dependency)

    @staticmethod
    def _find_dependency(publisher, subscriber, handler):
        dds = publisher.destruction_dependencies
        for dd in dds:
            if dd['handler'] != handler:
                continue
            d_subscriber = dd['subscriber']
            if d_subscriber:
                d_subscriber = d_subscriber()
            if d_subscriber == subscriber:
                return dd

    @staticmethod
    def collect():
        helpers.get_executor().object_store.cleanup()

    @staticmethod
    @specs.parameter('object_', dsl.MuranoObjectParameter(decorate=False))
    def is_doomed(object_):
        return helpers.get_object_store().is_doomed(object_)

    @staticmethod
    @specs.parameter('object_', dsl.MuranoObjectParameter(decorate=False))
    def is_destroyed(object_):
        return object_.destroyed



# Copyright 2016 AT&T Corp
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from oslo_config import cfg
from oslo_serialization import jsonutils
import ssl as ssl_module

from murano.common.i18n import _
from murano.common.messaging import mqclient
from murano.tests.unit import base

CONF = cfg.CONF


class MQClientTest(base.MuranoTestCase):

    @mock.patch('murano.common.messaging.mqclient.kombu')
    def setUp(self, mock_kombu):
        super(MQClientTest, self).setUp()
        self.ssl_client = mqclient.MqClient(login='test_login',
                                            password='test_password',
                                            host='test_host', port='test_port',
                                            virtual_host='test_virtual_host',
                                            ssl=True, ca_certs=['cert1'],
                                            insecure=False)

        mock_kombu.Connection.assert_called_once_with(
            'amqp://{0}:{1}@{2}:{3}/{4}'.format('test_login', 'test_password',
                                                'test_host', 'test_port',
                                                'test_virtual_host'),
            ssl={'ca_certs': ['cert1'], 'cert_reqs': ssl_module.CERT_REQUIRED})
        self.assertEqual(mock_kombu.Connection(), self.ssl_client._connection)
        self.assertIsNone(self.ssl_client._channel)
        self.assertFalse(self.ssl_client._connected)

    @mock.patch('murano.common.messaging.mqclient.kombu', autospec=True)
    def test_client_initialization_with_ssl_version(self, mock_kombu):
        ssl_versions = (
            ('tlsv1', getattr(ssl_module, 'PROTOCOL_TLSv1', None)),
            ('tlsv1_1', getattr(ssl_module, 'PROTOCOL_TLSv1_1', None)),
            ('tlsv1_2', getattr(ssl_module, 'PROTOCOL_TLSv1_2', None)),
            ('sslv2', getattr(ssl_module, 'PROTOCOL_SSLv2', None)),
            ('sslv23', getattr(ssl_module, 'PROTOCOL_SSLv23', None)),
            ('sslv3', getattr(ssl_module, 'PROTOCOL_SSLv3', None)))
        exception_count = 0

        for ssl_name, ssl_version in ssl_versions:
            ssl_kwargs = {
                'login': 'test_login',
                'password': 'test_password',
                'host': 'test_host',
                'port': 'test_port',
                'virtual_host': 'test_virtual_host',
                'ssl': True,
                'ssl_version': ssl_name,
                'ca_certs': ['cert1'],
                'insecure': False
            }

            # If a ssl_version is not valid, a RuntimeError is thrown.
            # According to the ssl_version docs in config.py, certain versions
            # of TLS may be available depending on the system. So, just
            # check that at least 1 ssl_version works.
            if ssl_version is None:
                e = self.assertRaises(RuntimeError, mqclient.MqClient,
                                      **ssl_kwargs)
                self.assertEqual(_('Invalid SSL version: %s') % ssl_name,
                                 e.__str__())
                exception_count += 1
                continue

            self.ssl_client = mqclient.MqClient(**ssl_kwargs)

            mock_kombu.Connection.assert_called_once_with(
                'amqp://{0}:{1}@{2}:{3}/{4}'.format(
                    'test_login', 'test_password', 'test_host', 'test_port',
                    'test_virtual_host'),
                ssl={'ca_certs': ['cert1'],
                     'cert_reqs': ssl_module.CERT_REQUIRED,
                     'ssl_version': ssl_version})
            self.assertEqual(
                mock_kombu.Connection(), self.ssl_client._connection)
            self.assertIsNone(self.ssl_client._channel)
            self.assertFalse(self.ssl_client._connected)
            mock_kombu.Connection.reset_mock()

        # Check that at least one ssl_version worked.
        self.assertGreater(len(ssl_versions), exception_count)

    @mock.patch('murano.common.messaging.mqclient.kombu')
    def test_alternate_client_initializations(self, mock_kombu):
        for ca_cert in ['cert1', None]:
            client = mqclient.MqClient(login='test_login',
                                       password='test_password',
                                       host='test_host', port='test_port',
                                       virtual_host='test_virtual_host',
                                       ssl=True,
                                       ca_certs=ca_cert,
                                       insecure=True)

            mock_kombu.Connection.assert_called_once_with(
                'amqp://{0}:{1}@{2}:{3}/{4}'.format('test_login',
                                                    'test_password',
                                                    'test_host', 'test_port',
                                                    'test_virtual_host'),
                ssl={'ca_certs': ca_cert,
                     'cert_reqs': ssl_module.CERT_OPTIONAL
                     if ca_cert else ssl_module.CERT_NONE})
            self.assertEqual(mock_kombu.Connection(),
                             client._connection)
            mock_kombu.Connection.reset_mock()

        client = mqclient.MqClient(login='test_login',
                                   password='test_password',
                                   host='test_host', port='test_port',
                                   virtual_host='test_virtual_host',
                                   ssl=False,
                                   ca_certs=None,
                                   insecure=False)

        mock_kombu.Connection.assert_called_once_with(
            'amqp://{0}:{1}@{2}:{3}/{4}'.format('test_login',
                                                'test_password',
                                                'test_host', 'test_port',
                                                'test_virtual_host'),
            ssl=None)
        self.assertEqual(mock_kombu.Connection(),
                         client._connection)

    def test_connect(self):
        self.ssl_client.connect()

        self.ssl_client._connection.connect.assert_called_once_with()
        self.ssl_client._connection.channel.assert_called_once_with()
        self.assertEqual(self.ssl_client._connection.channel(),
                         self.ssl_client._channel)
        self.assertTrue(self.ssl_client._connected)

    def test_close(self):
        self.ssl_client.close()

        self.ssl_client._connection.close.assert_called_once_with()
        self.assertFalse(self.ssl_client._connected)

    def test_enter_and_exit(self):
        with self.ssl_client:
            pass

        self.ssl_client._connection.connect.assert_called_once_with()
        self.ssl_client._connection.close.assert_called_once_with()

    @mock.patch('murano.common.messaging.mqclient.kombu')
    def test_declare(self, mock_kombu):
        self.ssl_client.connect()
        self.ssl_client.declare(queue='test_queue', exchange='test_exchange',
                                enable_ha=True, ttl=1)

        queue_args = {
            'x-ha-policy': 'all',
            'x-expires': 1
        }
        mock_kombu.Exchange.assert_called_once_with(
            'test_exchange', type='direct', durable=True)
        mock_kombu.Queue.assert_called_once_with('test_queue',
                                                 mock_kombu.Exchange(),
                                                 'test_queue',
                                                 durable=True,
                                                 queue_arguments=queue_args)
        mock_kombu.Queue()().declare.assert_called_once_with()
        mock_kombu.reset_mock()

        self.ssl_client.declare(queue='test_queue', exchange='test_exchange',
                                enable_ha=False, ttl=0)

        mock_kombu.Exchange.assert_called_once_with(
            'test_exchange', type='direct', durable=True)
        mock_kombu.Queue.assert_called_once_with('test_queue',
                                                 mock_kombu.Exchange(),
                                                 'test_queue',
                                                 durable=True,
                                                 queue_arguments={})

    def test_declare_except_runtime_error(self):
        with self.assertRaisesRegex(RuntimeError,
                                    'Not connected to RabbitMQ'):
            self.ssl_client.declare(None)

    @mock.patch('murano.common.messaging.mqclient.kombu')
    def test_send(self, mock_kombu):
        mock_message = mock.MagicMock(body='test_message', id=3)

        self.ssl_client.connect()
        self.ssl_client.send(mock_message, 'test_key', 'test_exchange')

        mock_kombu.Producer.assert_called_once_with(
            self.ssl_client._connection)
        mock_kombu.Producer().publish.assert_called_once_with(
            exchange='test_exchange', routing_key='test_key',
            body=jsonutils.dumps('test_message'), message_id='3',
            headers=None)

    @mock.patch('murano.common.messaging.mqclient.kombu')
    def test_send_signed(self, mock_kombu):
        mock_message = mock.MagicMock(body='test_message', id=3)

        signer = mock.MagicMock()
        signer.return_value = "SIGNATURE"
        self.ssl_client.connect()
        self.ssl_client.send(mock_message, 'test_key', 'test_exchange', signer)

        mock_kombu.Producer.assert_called_once_with(
            self.ssl_client._connection)
        mock_kombu.Producer().publish.assert_called_once_with(
            exchange='test_exchange', routing_key='test_key',
            body=jsonutils.dumps('test_message'), message_id='3',
            headers={'signature': 'SIGNATURE'})

    def test_send_except_runtime_error(self):
        with self.assertRaisesRegex(RuntimeError,
                                    'Not connected to RabbitMQ'):
            self.ssl_client.send(None, None)

    @mock.patch('murano.common.messaging.mqclient.subscription')
    def test_open(self, mock_subscription):
        self.ssl_client.connect()
        self.ssl_client.open('test_queue', prefetch_count=2)
        mock_subscription.Subscription.assert_called_once_with(
            self.ssl_client._connection, 'test_queue', 2)

    def test_open_except_runtime_error(self):
        with self.assertRaisesRegex(RuntimeError,
                                    'Not connected to RabbitMQ'):
            self.ssl_client.open(None)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""empty message

Revision ID: 001
Revises: None
Create Date: 2014-05-29 16:32:33.698760

"""
import uuid

from alembic import op
from oslo_utils import timeutils
import sqlalchemy as sa
from sqlalchemy.sql.expression import table as sa_table

from murano.common import consts
from murano.db.sqla import types as st

# revision identifiers, used by Alembic.
revision = '001'
down_revision = None

MYSQL_ENGINE = 'InnoDB'
MYSQL_CHARSET = 'utf8'


def _create_default_categories(op):
    bind = op.get_bind()
    table = sa_table(
        'category',
        sa.Column('id', sa.String(length=36), primary_key=True),
        sa.Column('created', sa.DateTime()),
        sa.Column('updated', sa.DateTime()),
        sa.Column('name', sa.String(length=80)))

    now = timeutils.utcnow()

    for category in consts.CATEGORIES:
        values = {'id': uuid.uuid4().hex,
                  'name': category,
                  'updated': now,
                  'created': now}
        bind.execute(table.insert(values=values))


def upgrade():
    op.create_table(
        'environment',
        sa.Column('created', sa.DateTime(), nullable=False),
        sa.Column('updated', sa.DateTime(), nullable=False),
        sa.Column('id', sa.String(length=255), nullable=False),
        sa.Column('name', sa.String(length=255), nullable=False),
        sa.Column('tenant_id', sa.String(length=36), nullable=False),
        sa.Column('version', sa.BigInteger(), nullable=False),
        sa.Column('description', sa.Text(), nullable=False),
        sa.Column('networking', sa.Text(), nullable=True),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('tenant_id', 'name'),
        mysql_engine=MYSQL_ENGINE,
        mysql_charset=MYSQL_CHARSET
    )

    op.create_table(
        'tag',
        sa.Column('created', sa.DateTime(), nullable=False),
        sa.Column('updated', sa.DateTime(), nullable=False),
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('name', sa.String(length=80), nullable=False),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('name'),
        mysql_engine=MYSQL_ENGINE,
        mysql_charset=MYSQL_CHARSET
    )

    op.create_table(
        'category',
        sa.Column('created', sa.DateTime(), nullable=False),
        sa.Column('updated', sa.DateTime(), nullable=False),
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('name', sa.String(length=80), nullable=False, unique=True),
        sa.PrimaryKeyConstraint('id'),
        mysql_engine=MYSQL_ENGINE,
        mysql_charset=MYSQL_CHARSET
    )

    op.create_index('ix_category_name', 'category', ['name'])

    op.create_table(
        'apistats',
        sa.Column('created', sa.DateTime(), nullable=False),
        sa.Column('updated', sa.DateTime(), nullable=False),
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('host', sa.String(length=80), nullable=True),
        sa.Column('request_count', sa.BigInteger(), nullable=True),
        sa.Column('error_count', sa.BigInteger(), nullable=True),
        sa.Column('average_response_time', sa.Float(), nullable=True),
        sa.Column('requests_per_tenant', sa.Text(), nullable=True),
        sa.Column('requests_per_second', sa.Float(), nullable=True),
        sa.Column('errors_per_second', sa.Float(), nullable=True),
        sa.Column('cpu_count', sa.Integer(), nullable=True),
        sa.Column('cpu_percent', sa.Float(), nullable=True),
        sa.PrimaryKeyConstraint('id'),
        mysql_engine=MYSQL_ENGINE,
        mysql_charset=MYSQL_CHARSET
    )

    op.create_table(
        'instance_stats',
        sa.Column('environment_id', sa.String(length=255), nullable=False),
        sa.Column('instance_id', sa.String(length=255), nullable=False),
        sa.Column('instance_type', sa.Integer(), nullable=False),
        sa.Column('created', sa.Integer(), nullable=False),
        sa.Column('destroyed', sa.Integer(), nullable=True),
        sa.Column('type_name', sa.String(length=512), nullable=False),
        sa.Column('type_title', sa.String(length=512), nullable=True),
        sa.Column('unit_count', sa.Integer(), nullable=True),
        sa.Column('tenant_id', sa.String(length=36), nullable=False),
        sa.PrimaryKeyConstraint('environment_id', 'instance_id'),
        mysql_engine=MYSQL_ENGINE,
        mysql_charset=MYSQL_CHARSET
    )

    op.create_table(
        'package',
        sa.Column('created', sa.DateTime(), nullable=False),
        sa.Column('updated', sa.DateTime(), nullable=False),
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('archive', st.LargeBinary(), nullable=True),
        sa.Column('fully_qualified_name',
                  sa.String(length=128),
                  nullable=False,
                  unique=True),
        sa.Column('type', sa.String(length=20), nullable=False),
        sa.Column('author', sa.String(length=80), nullable=True),
        sa.Column('name', sa.String(length=80), nullable=False),
        sa.Column('enabled', sa.Boolean(), nullable=True),
        sa.Column('description', sa.String(length=512), nullable=False),
        sa.Column('is_public', sa.Boolean(), nullable=True),
        sa.Column('logo', st.LargeBinary(), nullable=True),
        sa.Column('owner_id', sa.String(length=36), nullable=False),
        sa.Column('ui_definition', sa.Text(), nullable=True),
        sa.PrimaryKeyConstraint('id'),
        mysql_engine=MYSQL_ENGINE,
        mysql_charset=MYSQL_CHARSET
    )

    op.create_index('ix_package_fqn',
                    'package',
                    ['fully_qualified_name'])

    op.create_table(
        'session',
        sa.Column('created', sa.DateTime(), nullable=False),
        sa.Column('updated', sa.DateTime(), nullable=False),
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('environment_id', sa.String(length=255), nullable=True),
        sa.Column('user_id', sa.String(length=36), nullable=False),
        sa.Column('state', sa.String(length=36), nullable=False),
        sa.Column('description', sa.Text(), nullable=False),
        sa.Column('version', sa.BigInteger(), nullable=False),
        sa.ForeignKeyConstraint(['environment_id'], ['environment.id'], ),
        sa.PrimaryKeyConstraint('id'),
        mysql_engine=MYSQL_ENGINE,
        mysql_charset=MYSQL_CHARSET
    )

    op.create_table(
        'deployment',
        sa.Column('created', sa.DateTime(), nullable=False),
        sa.Column('updated', sa.DateTime(), nullable=False),
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('started', sa.DateTime(), nullable=False),
        sa.Column('finished', sa.DateTime(), nullable=True),
        sa.Column('description', sa.Text(), nullable=False),
        sa.Column('environment_id', sa.String(length=255), nullable=True),
        sa.ForeignKeyConstraint(['environment_id'], ['environment.id'], ),
        sa.PrimaryKeyConstraint('id'),
        mysql_engine=MYSQL_ENGINE,
        mysql_charset=MYSQL_CHARSET
    )

    op.create_table(
        'class_definition',
        sa.Column('created', sa.DateTime(), nullable=False),
        sa.Column('updated', sa.DateTime(), nullable=False),
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('name',
                  sa.String(length=128),
                  nullable=False,
                  unique=True),
        sa.Column('package_id', sa.String(length=36), nullable=True),
        sa.ForeignKeyConstraint(['package_id'], ['package.id'], ),
        sa.PrimaryKeyConstraint('id'),
        mysql_engine=MYSQL_ENGINE,
        mysql_charset=MYSQL_CHARSET
    )

    op.create_index('ix_class_definition_name',
                    'class_definition',
                    ['name'])

    op.create_table(
        'status',
        sa.Column('created', sa.DateTime(), nullable=False),
        sa.Column('updated', sa.DateTime(), nullable=False),
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('entity_id', sa.String(length=255), nullable=True),
        sa.Column('entity', sa.String(length=10), nullable=True),
        sa.Column('deployment_id', sa.String(length=36), nullable=True),
        sa.Column('text', sa.Text(), nullable=False),
        sa.Column('level', sa.String(length=32), nullable=False),
        sa.Column('details', sa.Text(), nullable=True),
        sa.ForeignKeyConstraint(['deployment_id'], ['deployment.id'], ),
        sa.PrimaryKeyConstraint('id'),
        mysql_engine=MYSQL_ENGINE,
        mysql_charset=MYSQL_CHARSET
    )

    op.create_table(
        'package_to_tag',
        sa.Column('package_id', sa.String(length=36), nullable=False),
        sa.Column('tag_id', sa.String(length=36), nullable=False),
        sa.ForeignKeyConstraint(['package_id'], ['package.id'], ),
        sa.ForeignKeyConstraint(['tag_id'], ['tag.id'], ondelete='CASCADE'),
        mysql_engine=MYSQL_ENGINE,
        mysql_charset=MYSQL_CHARSET
    )

    op.create_table(
        'package_to_category',
        sa.Column('package_id', sa.String(length=36), nullable=False),
        sa.Column('category_id', sa.String(length=36), nullable=False),
        sa.ForeignKeyConstraint(['category_id'],
                                ['category.id'],
                                ondelete='RESTRICT'),
        sa.ForeignKeyConstraint(['package_id'], ['package.id'], ),
        mysql_engine=MYSQL_ENGINE,
        mysql_charset=MYSQL_CHARSET
    )

    _create_default_categories(op)
    # end Alembic commands #


def downgrade():
    op.drop_index('ix_category_name', table_name='category')
    op.drop_index('ix_package_fqn', table_name='package')
    op.drop_index('ix_class_definition_name', table_name='class_definition')

    op.drop_table('status')
    op.drop_table('package_to_category')
    op.drop_table('class_definition')
    op.drop_table('deployment')
    op.drop_table('package_to_tag')
    op.drop_table('session')
    op.drop_table('instance_stats')
    op.drop_table('package')
    op.drop_table('apistats')
    op.drop_table('category')
    op.drop_table('tag')
    op.drop_table('environment')
    # end Alembic commands #



# Copyright (c) 2016 AT&T
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from murano.api.v1 import instance_statistics
import murano.tests.unit.api.base as tb


class TestInstanceStatistics(tb.ControllerTest, tb.MuranoApiTestCase):
    def setUp(self):
        super(TestInstanceStatistics, self).setUp()
        self.controller = instance_statistics.Controller()

    def test_get_aggregated(self):
        self._set_policy_rules(
            {"get_aggregated_statistics": "@"}
        )
        self.expect_policy_check("get_aggregated_statistics",
                                 {'environment_id': u'12344'})
        env_id = 12344

        req = self._get('/environments/{env_id}/'
                        'instance-statistics/aggregated'
                        .format(env_id=env_id))
        result = req.get_response(self.api)
        self.assertEqual(200, result.status_code)

    def test_get_for_instance(self):
        self._set_policy_rules(
            {"get_instance_statistics": "@"}
        )
        self.expect_policy_check("get_instance_statistics",
                                 {'environment_id': u'12345',
                                  'instance_id': u'12'})
        env_id = 12345
        ins_id = 12

        req = self._get('/environments/{env_id}/'
                        'instance-statistics/raw/{ins_id}'
                        .format(env_id=env_id, ins_id=ins_id))
        result = req.get_response(self.api)
        self.assertEqual(200, result.status_code)

    def test_get_for_env(self):
        self._set_policy_rules(
            {"get_statistics": "@"}
        )
        self.expect_policy_check("get_statistics",
                                 {'environment_id': u'12346'})
        env_id = 12346

        req = self._get('/environments/{env_id}/instance-statistics/raw'
                        .format(env_id=env_id))
        result = req.get_response(self.api)
        self.assertEqual(200, result.status_code)



# Copyright (c) 2016 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from murano.dsl.contracts import contracts
from murano.dsl import dsl_types
from murano.dsl import executor
from murano.dsl import helpers
from murano.dsl import meta as meta_module
from murano.dsl import murano_type


def generate_schema(pkg_loader, context_manager,
                    class_name, method_names=None,
                    class_version=None, package_name=None):
    """Generate JSON schema

    JSON Schema is generated either for the class with all model builders
    or for specified model builders only. The return value is a dictionary
    with keys being model builder names and the values are JSON schemas for
    them. The class itself is represented by an empty string key.
    """

    if method_names and not isinstance(method_names, (list, tuple)):
        method_names = (method_names,)
    version = helpers.parse_version_spec(class_version)
    if package_name:
        package = pkg_loader.load_package(package_name, version)
    else:
        package = pkg_loader.load_class_package(class_name, version)

    cls = package.find_class(class_name, search_requirements=False)
    exc = executor.MuranoDslExecutor(pkg_loader, context_manager)
    with helpers.with_object_store(exc.object_store):
        context = exc.create_object_context(cls)
        model_builders = set(list_model_builders(cls, context))
        method_names = model_builders.intersection(
            method_names or model_builders)

        result = {
            name: generate_entity_schema(
                get_entity(cls, name), context, cls,
                get_meta(cls, name, context))
            for name in method_names
        }
        return result


def list_model_builders(cls, context):
    """List model builder names of the class

    Yield names of all model builders (static actions marked with appropriate
    metadata) plus empty string for the class itself.
    """

    yield ''
    for method_name in cls.all_method_names:
        try:
            method = cls.find_single_method(method_name)
            if not method.is_action or not method.is_static:
                continue
            meta = meta_module.aggregate_meta(method, context)
            is_builder = meta.get('io.murano.metadata.ModelBuilder')
            if is_builder and is_builder.get_property('enabled'):
                yield method.name
        except Exception:
            pass


def get_meta(cls, method_name, context):
    """Get metadata dictionary for the method or class"""
    if not method_name:
        return meta_module.aggregate_meta(cls, context)

    method = cls.find_single_method(method_name)
    return meta_module.aggregate_meta(method, context)


def get_entity(cls, method_name):
    """Get MuranoMethod of the class by its name"""
    if not method_name:
        return cls
    method = cls.find_single_method(method_name)
    return method


def get_properties(entity):
    """Get properties/arg scheme of the class/method"""
    if isinstance(entity, dsl_types.MuranoType):
        properties = entity.all_property_names
        result = {}
        for prop_name in properties:
            prop = entity.find_single_property(prop_name)

            if prop.usage not in (dsl_types.PropertyUsages.In,
                                  dsl_types.PropertyUsages.InOut):
                continue
            result[prop_name] = prop
        return result
    return entity.arguments_scheme


def generate_entity_schema(entity, context, declaring_type, meta):
    """Generate schema for single class or method by it DSL entity"""
    properties = get_properties(entity)
    type_weights = murano_type.weigh_type_hierarchy(declaring_type)
    schema = {
        '$schema': 'http://json-schema.org/draft-04/schema#',
        'type': 'object',
        'properties': {
            name: generate_property_schema(prop, context, type_weights)
            for name, prop in properties.items()
            },
        'additionalProperties': False,
        'formSections': generate_sections(meta, type_weights)
    }
    schema.update(generate_ui_hints(entity, context, type_weights))
    return schema


def generate_sections(meta, type_weights):
    """Builds sections definitions for the schema

    Sections are UI hint for UI for grouping inputs into tabs/group-boxes.
    The code collects section definitions from type hierarchy considering that
    the section might be redefined in ancestor with the different section
    index and then re-enumerates them in a way that sections from the most
    base classes in hierarchy will get lower index values and there be no
    two sections with the same index.
    """
    section_list = meta.get('io.murano.metadata.forms.Section', [])
    sections_map = {}
    for section in section_list:
        name = section.get_property('name')
        ex_section = sections_map.get(name)
        if not ex_section:
            pass
        elif (type_weights[ex_section.declaring_type.name] <
                type_weights[section.declaring_type.name]):
            continue
        elif (type_weights[ex_section.declaring_type.name] ==
                type_weights[section.declaring_type.name]):
            index = section.get_property('index')
            if index is None:
                continue
            ex_index = ex_section.get_property('index')
            if ex_index is not None and ex_index <= index:
                continue
        sections_map[name] = section

    ordered_sections, unordered_sections = sort_by_index(
        sections_map.values(), type_weights)
    sections = {}
    index = 0
    for section in ordered_sections:
        name = section.get_property('name')
        if name not in sections:
            sections[name] = {
                'title': section.get_property('title'),
                'index': index
            }
            index += 1
    for section in unordered_sections:
        name = section.get_property('name')
        if name not in sections:
            sections[name] = {
                'title': section.get_property('title')
            }
    return sections


def generate_property_schema(prop, context, type_weights):
    """Generate schema for single property/argument"""
    schema = translate(prop.contract.spec, context,
                       prop.declaring_type.package.runtime_version)
    if prop.has_default:
        schema['default'] = prop.default
    schema.update(generate_ui_hints(prop, context, type_weights))
    return schema


def generate_ui_hints(entity, context, type_weights):
    """Translate know property/arg meta into json-schema UI hints"""
    schema = {}
    meta = meta_module.aggregate_meta(entity, context)
    for cls_name, schema_prop, meta_prop in (
            ('io.murano.metadata.Title', 'title', 'text'),
            ('io.murano.metadata.Description', 'description', 'text'),
            ('io.murano.metadata.HelpText', 'helpText', 'text'),
            ('io.murano.metadata.forms.Hidden', 'visible', 'visible')):
        value = meta.get(cls_name)
        if value is not None:
            schema[schema_prop] = value.get_property(meta_prop)

    position = meta.get('io.murano.metadata.forms.Position')
    if position:
        schema['formSection'] = position.get_property('section')
        index = position.get_property('index')
        if index is not None:
            schema['formIndex'] = (
                (position.get_property('index') + 1) * len(type_weights) -
                type_weights[position.declaring_type.name])

    return schema


def sort_by_index(meta, type_weights, property_name='index'):
    """Sorts meta definitions by its distance in the class hierarchy"""
    has_index = filter(
        lambda m: m.get_property(property_name) is not None, meta)
    has_no_index = filter(
        lambda m: m.get_property(property_name) is None, meta)

    return (
        sorted(has_index,
               key=lambda m: (
                   (m.get_property(property_name) + 1) *
                   len(type_weights) -
                   type_weights[m.declaring_type.name])),
        has_no_index)


def translate(contract, context, runtime_version):
    """Translates contracts into json-schema equivalents"""
    if isinstance(contract, dict):
        return translate_dict(contract, context, runtime_version)
    elif isinstance(contract, list):
        return translate_list(contract, context, runtime_version)
    elif isinstance(contract, dsl_types.YaqlExpression):
        return contracts.Contract.generate_expression_schema(
            contract, context, runtime_version)


def translate_dict(contract, context, runtime_version):
    """Translates dictionary contracts into json-schema objects"""
    properties = {}
    additional_properties = False
    for key, value in contract.items():
        if isinstance(key, dsl_types.YaqlExpression):
            additional_properties = translate(value, context, runtime_version)
        else:
            properties[key] = translate(value, context, runtime_version)
    return {
        'type': 'object',
        'properties': properties,
        'additionalProperties': additional_properties
    }


def translate_list(contract, context, runtime_version):
    """Translates list contracts into json-schema arrays"""
    items = []
    for value in contract:
        if isinstance(value, int):
            pass
        else:
            items.append(translate(value, context, runtime_version))
    if len(items) == 0:
        return {'type': 'array'}
    elif len(items) == 1:
        return {
            'type': 'array',
            'items': items[0],
        }
    else:
        return {
            'type': 'array',
            'items': items,
            'additionalItems': items[-1]
        }



#    Copyright (c) 2014 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os.path


from murano.dsl import constants
from murano.dsl import murano_package
from murano.dsl import namespace_resolver
from murano.dsl import package_loader
from murano.engine import yaql_yaml_loader
from murano.tests.unit.dsl.foundation import object_model
from oslo_utils import fnmatch


class TestPackage(murano_package.MuranoPackage):
    def __init__(self, pkg_loader, name, version,
                 runtime_version, requirements, configs, meta):
        self.__configs = configs
        super(TestPackage, self).__init__(
            pkg_loader, name, version,
            runtime_version, requirements, meta)

    def get_class_config(self, name):
        return self.__configs.get(name, {})

    def get_resource(self, name):
        pass


class TestPackageLoader(package_loader.MuranoPackageLoader):
    _classes_cache = {}

    def __init__(self, directory, package_name, parent_loader=None, meta=None):
        self._package_name = package_name
        self._yaml_loader = yaql_yaml_loader.get_loader('1.0')
        if directory in TestPackageLoader._classes_cache:
            self._classes = TestPackageLoader._classes_cache[directory]
        else:
            self._classes = {}
            self._build_index(directory)
            TestPackageLoader._classes_cache[directory] = self._classes
        self._parent = parent_loader
        self._configs = {}
        self._package = TestPackage(
            self, package_name, None, constants.RUNTIME_VERSION_1_0,
            None, self._configs, meta)
        for name, payload in self._classes.items():
            self._package.register_class(payload, name)
        super(TestPackageLoader, self).__init__()

    def load_package(self, package_name, version_spec):
        if package_name == self._package_name:
            return self._package
        elif self._parent:
            return self._parent.load_package(package_name, version_spec)
        else:
            raise KeyError(package_name)

    def load_class_package(self, class_name, version_spec):
        if class_name in self._classes:
            return self._package
        elif self._parent:
            return self._parent.load_class_package(class_name, version_spec)
        else:
            raise KeyError(class_name)

    def export_fixation_table(self):
        return {}

    def import_fixation_table(self, fixations):
        pass

    def compact_fixation_table(self):
        pass

    def _build_index(self, directory):
        yamls = [
            os.path.join(dirpath, f)
            for dirpath, _, files in os.walk(directory)
            for f in fnmatch.filter(files, '*.yaml')
            if f != 'manifest.yaml'
        ]
        for class_def_file in yamls:
            self._load_classes(class_def_file)

    def _load_classes(self, class_def_file):
        with open(class_def_file, 'rb') as stream:
            data_lst = self._yaml_loader(stream.read(), class_def_file)

        last_ns = {}
        for data in data_lst:
            last_ns = data.get('Namespaces', last_ns.copy())
            if 'Name' not in data:
                continue

            for name, method in (data.get('Methods') or data.get(
                    'Workflow') or {}).items():
                if name.startswith('test'):
                    method['Scope'] = 'Public'

            ns = namespace_resolver.NamespaceResolver(last_ns)
            class_name = ns.resolve_name(data['Name'])
            self._classes[class_name] = data_lst

    def set_config_value(self, class_name, property_name, value):
        if isinstance(class_name, object_model.Object):
            class_name = class_name.type_name
        self._configs.setdefault(class_name, {})[
            property_name] = value

    def register_package(self, package):
        super(TestPackageLoader, self).register_package(package)



# Copyright 2016 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Increase time resolution for status reports

Revision ID: 014
Create Date: 2016-04-28

"""

from alembic import op
import sqlalchemy.dialects.mysql as sa_mysql

# revision identifiers, used by Alembic.
revision = '014'
down_revision = '013'

MYSQL_ENGINE = 'InnoDB'
MYSQL_CHARSET = 'utf8'


def _check_dbms(engine):
    dialect = engine.dialect.dialect_description
    version = engine.dialect.server_version_info
    if dialect.startswith('mysql') and version >= (5, 6, 4):
        return True
    if 'MariaDB' in version and version >= (5, 3):
        return True
    return False


def upgrade():
    engine = op.get_bind()
    if _check_dbms(engine):
        with op.batch_alter_table('status') as batch_op:
            batch_op.alter_column(
                'created', type_=sa_mysql.DATETIME(fsp=6), nullable=False)
            batch_op.alter_column(
                'updated', type_=sa_mysql.DATETIME(fsp=6), nullable=False)


def downgrade():
    engine = op.get_bind()
    if _check_dbms(engine):
        with op.batch_alter_table('status') as batch_op:
            batch_op.alter_column(
                'created', type_=sa_mysql.DATETIME(), nullable=False)
            batch_op.alter_column(
                'updated', type_=sa_mysql.DATETIME(), nullable=False)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import yaml

from murano.common.helpers import path
from murano.packages import exceptions
from murano.packages import package_base

RESOURCES_DIR_NAME = 'Resources/'


class YAQL(object):
    def __init__(self, expr):
        self.expr = expr


class Dumper(yaml.SafeDumper):
    pass


def yaql_representer(dumper, data):
    return dumper.represent_scalar(u'!yaql', data.expr)


Dumper.add_representer(YAQL, yaql_representer)


class CloudifyToscaPackage(package_base.PackageBase):
    def __init__(self, format_name, runtime_version, source_directory,
                 manifest):
        super(CloudifyToscaPackage, self).__init__(
            format_name, runtime_version, source_directory, manifest)

        self._entry_point = manifest.get('EntryPoint', 'main.yaml')
        self._generated_class = None
        self._generated_ui = None

    @property
    def classes(self):
        return self.full_name,

    @property
    def requirements(self):
        return {
            'org.getcloudify.murano': '0'
        }

    @property
    def ui(self):
        if not self._generated_ui:
            self._generated_ui = self._generate_ui()
        return self._generated_ui

    def get_class(self, name):
        if name != self.full_name:
            raise exceptions.PackageClassLoadError(
                name, 'Class not defined in this package')
        if not self._generated_class:
            self._generated_class = self._generate_class()
        return self._generated_class, '<generated code>'

    def _generate_class(self):
        inputs, outputs = self._get_inputs_outputs()
        class_code = {
            'Name': self.full_name,
            'Extends': 'org.getcloudify.murano.CloudifyApplication',
            'Properties': self._generate_properties(inputs, outputs),
            'Methods': {
                'describe': self._generate_describe_method(inputs),
                'updateOutputs': self._generate_update_outputs_method(outputs)
            }
        }
        return yaml.dump(class_code, Dumper=Dumper, default_style='"')

    @staticmethod
    def _generate_properties(inputs, outputs):
        contracts = {}
        for name, value in inputs.items():
            prop = {
                'Contract': YAQL('$.string().notNull()'),
                'Usage': 'In'
            }
            if 'default' in value:
                prop['Default'] = value['default']
            contracts[name] = prop

        for name in outputs.keys():
            contracts[name] = {
                'Contract': YAQL('$.string()'),
                'Usage': 'Out'
            }

        return contracts

    def _generate_describe_method(self, inputs):
        input_values = {
            name: YAQL('$.' + name)
            for name in inputs.keys()
        }

        return {
            'Body': [{
                'Return': {
                    'entryPoint': self._entry_point,
                    'inputs': input_values
                }
            }]
        }

    @staticmethod
    def _generate_update_outputs_method(outputs):
        assignments = [
            {YAQL('$.' + name): YAQL('$outputs.get({0})'.format(name))}
            for name in outputs.keys()
        ]
        return {
            'Arguments': [{
                'outputs': {
                    'Contract': {
                        YAQL('$.string().notNull()'): YAQL('$')
                    }
                }
            }],
            'Body': assignments
        }

    def _get_inputs_outputs(self):
        entry_point_path = path.secure_join(
            self.source_directory, RESOURCES_DIR_NAME, self._entry_point)
        with open(entry_point_path) as blueprint:
            data = yaml.safe_load(blueprint)
            return data.get('inputs') or {}, data.get('outputs') or {}

    def _generate_application_ui_section(self, inputs, package_name=None,
                                         package_version=None):
        section = {
            key: YAQL(
                '$.appConfiguration.' + key) for key in inputs.keys()
        }
        section.update({
            '?': {
                'type': self.full_name
            }
        })
        if package_name:
            section['?']['package'] = package_name
        if package_version:
            section['?']['classVersion'] = package_version
        return section

    @staticmethod
    def _generate_form_ui_section(inputs):
        fields = [
            {
                'name': key,
                'label': key.title().replace('_', ' '),
                'type': 'string',
                'required': True,
                'description': value.get('description', key)
            } for key, value in inputs.items()
        ]
        return [{
            'appConfiguration': {
                'fields': fields
            }
        }]

    def _generate_ui(self):
        inputs, outputs = self._get_inputs_outputs()
        ui = {
            'Version': '2.2',
            'Application': self._generate_application_ui_section(
                inputs, self.full_name, str(self.version)),
            'Forms': self._generate_form_ui_section(inputs)
        }
        return yaml.dump(ui, Dumper=Dumper, default_style='"')



#    Copyright (c) 2016 AT&T
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import testtools
from unittest import mock

from murano.dsl import murano_method
from murano.dsl import murano_type
from murano.dsl import object_store
from murano.engine.system import test_fixture
from murano.tests.unit import base


class TestTestFixture(base.MuranoTestCase):
    def setUp(self):
        super(TestTestFixture, self).setUp()

        self.mock_class = mock.MagicMock(spec=murano_type.MuranoClass)
        self.mock_method = mock.MagicMock(spec=murano_method.MuranoMethod)
        self.mock_object_store = mock.Mock(spec=object_store.ObjectStore)

        self.test_fixture = test_fixture.TestFixture()

        self.addCleanup(mock.patch.stopall)

    @mock.patch("murano.dsl.helpers.get_execution_session")
    def test_finish_env(self, execution_session):
        self.assertIsNone(self.test_fixture.finish_env())
        self.assertTrue(execution_session.called)

    @mock.patch("murano.dsl.helpers.get_execution_session")
    def test_start_env(self, execution_session):
        self.assertIsNone(self.test_fixture.start_env())
        self.assertTrue(execution_session.called)

    @mock.patch("murano.dsl.helpers.get_executor")
    def test_load(self, executor):
        executor.return_value = self.mock_object_store
        model = "test"
        tf_load = self.test_fixture.load(model)
        self.assertEqual(self.test_fixture.load(model), tf_load)

    def test_assert_true(self):
        expr = (7 > 3)
        message = None
        # Calls assertTrue in super class
        self.test_fixture.assert_true(expr, message)

    def test_assert_false(self):
        expr = (3 != 3)
        message = None
        # Calls assertFalse in super class
        self.test_fixture.assert_false(expr, message)

    def test_assert_in(self):
        needle = 7
        haystack = [3, 7, 8, 9, 22]
        message = None
        # Calls assertIn in super class
        self.test_fixture.assert_in(needle, haystack, message)

    def test_assert_not_in(self):
        needle = 16
        haystack = [3, 7, 8, 9, 22]
        message = None
        # Calls assertNotIn in super class
        self.test_fixture.assert_not_in(needle, haystack, message)

    def test_assert_true_fails(self):
        expr = (7 < 3)
        message = None
        self.assertRaises(AssertionError,
                          self.test_fixture.assert_true, expr, message)

    def test_assert_false_fails(self):
        expr = (3 == 3)
        message = None
        self.assertRaises(AssertionError,
                          self.test_fixture.assert_false, expr, message)

    def test_assert_in_fails(self):
        needle = 25
        haystack = [3, 7, 8, 9, 22]
        message = None
        self.assertRaises(testtools.matchers._impl.MismatchError,
                          self.test_fixture.assert_in, needle,
                          haystack, message)

    def test_assert_not_in_fails(self):
        needle = 22
        haystack = [3, 7, 8, 9, 22]
        message = None
        self.assertRaises(testtools.matchers._impl.MismatchError,
                          self.test_fixture.assert_not_in, needle,
                          haystack, message)



# Copyright 2017 AT&T Corporation.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_policy import policy

from murano.common.policies import base

action_policies = [
    policy.DocumentedRuleDefault(
        name='execute_action',
        check_str=base.RULE_DEFAULT,
        description="""Excute an available action on a deployed environment,
retrieve the task status of an executed action, or retrieve the result of
an executed static action.""",
        operations=[
            {'path': 'v1/environments/{environment_id}/actions/{action_id}',
             'method': 'POST'},
            {'path': 'v1/environments/{environment_id}/actions/{task_id}',
             'method': 'GET'},
            {'path': 'v1/actions',
             'method': 'POST'}])
]


def list_rules():
    return action_policies



#    Copyright (c) 2013 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
import json

from murano.common import utils

from murano.tests.unit import base


class UtilsTests(base.MuranoTestCase):

    def test_validate_quotes(self):
        self.assertTrue(utils.validate_quotes('"ab"'))

    def test_validate_quotes_not_closed_quotes(self):
        self.assertRaises(ValueError, utils.validate_quotes, '"ab","b""')

    def test_validate_quotes_not_opened_quotes(self):
        self.assertRaises(ValueError, utils.validate_quotes, '""ab","b"')

    def test_validate_quotes_no_coma_before_opening_quotes(self):
        self.assertRaises(ValueError, utils.validate_quotes, '"ab""b"')

    def test_split_for_quotes(self):
        self.assertEqual(["a,b", "ac"], utils.split_for_quotes('"a,b","ac"'))

    def test_split_for_quotes_with_backslash(self):
        self.assertEqual(['a"bc', 'de', 'fg,h', r'klm\\', '"nop'],
                         utils.split_for_quotes(r'"a\"bc","de",'
                                                r'"fg,h","klm\\","\"nop"'))

    def test_validate_body(self):
        json_schema = json.dumps(['foo', {'bar': ('baz', None, 1.0, 2)}])
        self.assertIsNotNone(utils.validate_body(json_schema))

        json_schema = json.dumps(['body', {'body': ('baz', None, 1.0, 2)}])
        self.assertIsNotNone(utils.validate_body(json_schema))

    def test_build_entity_map(self):
        entity = {"?": {"fun": "id"}}
        self.assertEqual({}, utils.build_entity_map(entity))

        entity = {"?": {"id": "id"}}
        self.assertEqual({'id': {'?': {'id': 'id'}}},
                         utils.build_entity_map(entity))

        entity = [{"?": {"id": "id1"}}, {"?": {"id": "id2"}}]
        self.assertEqual({'id1': {'?': {'id': 'id1'}},
                          'id2': {'?': {'id': 'id2'}}},
                         utils.build_entity_map(entity))

    def test_is_different(self):
        t1 = "Hello"
        t2 = "World"
        self.assertTrue(utils.is_different(t1, t2))

        t1 = "Hello"
        t2 = "Hello"
        self.assertFalse(utils.is_different(t1, t2))

        t1 = {1, 2, 3, 4}
        t2 = t1
        self.assertFalse(utils.is_different(t1, t2))

        t2 = {1, 2, 3}
        self.assertTrue(utils.is_different(t1, t2))

        t1 = [1, 2, {1, 2, 3, 4}]
        t1[0] = t1
        self.assertTrue(utils.is_different(t1, t2))

        t1 = [t2]
        t2 = [t1]
        self.assertTrue(utils.is_different(t1, t2))

        t1 = [{1, 2, 3}, {1, 2, 3}]
        t2 = [{1, 2, 3}, {1, 2}]
        self.assertTrue(utils.is_different(t1, t2))

        t1 = datetime.date(2016, 8, 8)
        t2 = datetime.date(2016, 8, 7)
        self.assertTrue(utils.is_different(t1, t2))

        t1 = {1: 1, 2: 2, 3: 3}
        t2 = {1: 1, 2: 4, 3: 3}
        self.assertTrue(utils.is_different(t1, t2))

        t1 = {1: 1, 2: 2, 3: 3, 4: {"a": "hello", "b": [1, 2, 3]}}
        t2 = {1: 1, 2: 2, 3: 3, 4: {"a": "hello", "b": "world\n\n\nEnd"}}
        self.assertTrue(utils.is_different(t1, t2))

        t1 = {1: 1, 2: 2, 3: 3, 4: {"a": "hello", "b": [1, 2, 5]}}
        t2 = {1: 1, 2: 2, 3: 3, 4: {"a": "hello", "b": [1, 3, 2, 5]}}
        self.assertTrue(utils.is_different(t1, t2))

        class ClassA(object):
            __slots__ = ['x', 'y']

            def __init__(self, x, y):
                self.x = x
                self.y = y

        t1 = ClassA(1, 1)
        t2 = ClassA(1, 2)
        self.assertTrue(utils.is_different(t1, t2))

        t1 = [1, 2, 3]
        t1.append(t1)

        t2 = [1, 2, 4]
        t2.append(t2)
        self.assertTrue(utils.is_different(t1, t2))

        t1 = [1, 2, 3]
        t2 = [1, 2, 4]
        t2.append(t1)
        t1.append(t2)
        self.assertTrue(utils.is_different(t1, t2))

        t1 = utils
        t2 = datetime
        self.assertTrue(utils.is_different(t1, t2))

        t2 = "Not a module"
        self.assertTrue(utils.is_different(t1, t2))



# Copyright (c) 2014 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import shutil
import yaml
import zipfile

from murano import context
from murano.db import session


MANIFEST = {'Format': 'MuranoPL/1.0',
            'Type': 'Application',
            'Description': 'MockApp for API tests',
            'Author': 'Mirantis, Inc'}


def dummy_context(user='test_username', tenant_id='test_tenant_id',
                  request_id='dummy-request', **kwargs):

    # NOTE(kzaitsev) we need to pass non-False value to request_id, to
    # prevent it being generated by oslo during tests.
    params = {
        'request_id': request_id,
        'tenant': tenant_id,
        'user': user,
    }
    params.update(kwargs)
    return context.RequestContext.from_dict(params)


def save_models(*models):
    s = session.get_session()
    for m in models:
        m.save(s)


def compose_package(app_name, package_dir,
                    require=None, archive_dir=None, add_class_name=False,
                    manifest_required=True, version=None):
    """Composes a murano package

    Composes package `app_name` manifest and files from `package_dir`.
    Includes `require` section if any in the manifest file.
    Puts the resulting .zip file into `archive_dir` if present or in the
    `package_dir`.
    """
    tmp_package_dir = os.path.join(archive_dir, os.path.basename(package_dir))
    shutil.copytree(package_dir, tmp_package_dir)
    package_dir = tmp_package_dir

    if manifest_required:
        manifest = os.path.join(package_dir, "manifest.yaml")
        with open(manifest, 'w') as f:
            fqn = 'io.murano.apps.' + app_name
            mfest_copy = MANIFEST.copy()
            mfest_copy['FullName'] = fqn
            mfest_copy['Name'] = app_name
            mfest_copy['Classes'] = {fqn: 'mock_muranopl.yaml'}
            if require:
                mfest_copy['Require'] = {str(name): version
                                         for name, version in require}
            if version:
                mfest_copy['Version'] = version
            f.write(yaml.dump(mfest_copy, default_flow_style=False))

    if add_class_name:
        class_file = os.path.join(package_dir, 'Classes', 'mock_muranopl.yaml')
        with open(class_file, 'r') as f:
            contents = f.read()

        index = contents.index('Extends')
        contents = "{0}Name: {1}\n\n{2}".format(contents[:index], app_name,
                                                contents[index:])
        with open(class_file, 'w') as f:
            f.write(contents)

    if require:
        class_file = os.path.join(package_dir, 'Classes', 'mock_muranopl.yaml')
        with open(class_file, 'r') as f:
            content = f.read()

        index_string = 'deploy:\n    Body:\n      '
        index = content.index(index_string) + len(index_string)
        class_names = [req[0][req[0].rfind('.') + 1:] for req in require]
        addition = "".join(["- new({})\n".format(name) + ' ' * 6
                            for name in class_names])
        content = content[:index] + addition + content[index:]
        with open(class_file, 'w') as f:
            f.write(content)

    name = app_name + '.zip'

    if not archive_dir:
        archive_dir = os.path.dirname(os.path.abspath(__file__))
    archive_path = os.path.join(archive_dir, name)

    with zipfile.ZipFile(archive_path, 'w') as zip_file:
        for root, dirs, files in os.walk(package_dir):
            for f in files:
                zip_file.write(
                    os.path.join(root, f),
                    arcname=os.path.join(os.path.relpath(root, package_dir), f)
                )

    return archive_path, name



# Copyright (c) 2015 OpenStack Foundation.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging
from oslo_utils import importutils
from stevedore import extension
import yaml

LOG = logging.getLogger(__name__)


class ModifyActionManager(object):
    """Manages modify actions

    The manager encapsulates extensible plugin mechanism for
    modify actions loading. Provides ability to apply action on
    given object model based on action specification retrieved
    from congress
    """

    def __init__(self):
        self._cache = {}

    def load_action(self, name):
        """Loads action by its name

        Loaded actions are cached. Plugin mechanism is based on
        distutils entry points. Entry point namespace is
        'murano_policy_modify_actions'

        :param name: action name
        :return:
        """
        if name in self._cache:
            return self._cache[name]
        action = self._load_action(name)
        self._cache[name] = action
        return action

    @staticmethod
    def _load_action(name):
        mgr = extension.ExtensionManager(
            namespace='murano_policy_modify_actions',
            invoke_on_load=False
        )
        for ext in mgr.extensions:
            if name == ext.name:
                target = ext.entry_point_target.replace(':', '.')
                return importutils.import_class(target)
        raise ValueError('No such action definition: {action_name}'
                         .format(action_name=name))

    def apply_action(self, obj, action_spec):
        """Apply action on given model

        Parse action and its parameters from action specification
        retrieved from congress. Action specification is YAML format.

        E.g. remove-object: {object_id: abc123}")

        Action names are keys in top-level dictionary. Values are
        dictionaries containing key/value parameters of the action

        :param obj: subject of modification
        :param action_spec: YAML action spec
        :raise ValueError: in case of malformed action spec
        """
        actions = yaml.safe_load(action_spec)
        if not isinstance(actions, dict):
            raise ValueError('Expected action spec format is '
                             '"action-name: {{p1: v1, ...}}" '
                             'but got "{action_spec}"'
                             .format(action_spec=action_spec))
        for name, kwargs in actions.items():
            LOG.debug('Executing action {name}, params {params}'
                      .format(name=name, params=kwargs))
            # loads action class
            action_class = self.load_action(name)
            # creates action instance
            action_instance = action_class(**kwargs)
            # apply action on object model
            action_instance.modify(obj)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import argparse
import os
import re
import sys
import traceback

import eventlet
from keystoneclient.v3 import client as ks_client
from muranoclient.common import exceptions as exc
from muranoclient.common import utils
from oslo_config import cfg
from oslo_db import options
from oslo_log import log as logging
from oslo_utils import timeutils

from murano.common import config
from murano.common import engine
from murano.common.i18n import _
from murano.dsl import dsl_exception
from murano.dsl import dsl_types
from murano.dsl import exceptions
from murano.dsl import executor
from murano.dsl import helpers
from murano.engine import execution_session
from murano.engine import mock_context_manager
from murano.engine import package_loader
from murano import version


CONF = cfg.CONF
LOG = logging.getLogger(__name__)
options.set_defaults(CONF)

BASE_CLASS = 'io.murano.test.TestFixture'
TEST_CASE_NAME = re.compile('^test(?![a-z])')

OK_COLOR = '\033[92m'
FAIL_COLOR = '\033[91m'
END_COLOR = '\033[0m'


if os.name == 'nt':
    # eventlet monkey patching causes subprocess.Popen to fail on Windows
    # when using pipes due to missing non blocking I/O support
    eventlet.monkey_patch(os=False)
else:
    eventlet.monkey_patch()


class MuranoTestRunner(object):
    def __init__(self):
        self.parser = self.get_parser()
        self.args = self.parser.parse_args()

        if self.args.verbose:
            LOG.logger.setLevel(logging.DEBUG)

    def _load_package(self, pkg_loader, name):
        try:
            parts = name.rsplit('/')
            if len(parts) == 2:
                name, pkg_version = parts
                version_spec = helpers.parse_version_spec(pkg_version)
            else:
                version_spec = helpers.parse_version_spec('*')
            package = pkg_loader.load_package(name, version_spec)
        except exceptions.NoPackageFound:
            if not CONF.engine.load_packages_from:
                msg = _('Local package is not found since "load-packages-from"'
                        ' engine parameter is not provided and specified '
                        'packages is not loaded to murano-api')
            else:
                msg = _('Specified package is not found: {0} were scanned '
                        'together with murano database'
                        ).format(','.join(
                            CONF.engine.load_packages_from))
            LOG.error(msg)
            self.error(msg, show_help=False)
        except exc.CommunicationError:
            msg = ('Murano API is not running. '
                   'Check configuration parameters.')
            LOG.error(msg)
            self.error(msg, show_help=False)
        return package

    def _get_methods_to_run(self, package, tests_to_run, class_to_methods):
        if not tests_to_run:
            return class_to_methods

        methods_to_run = {}

        for item in tests_to_run:
            # Check for method name occurrence in all methods.
            # if there is no dot in provided item - it is a method name
            if '.' not in item:
                for class_name, methods in class_to_methods.items():
                    methods_to_run[class_name] = []
                    if item in methods:
                        methods_to_run[class_name].append(item)
                continue
            # Check match for the whole class name
            if item in package.classes:
                # Add all test cases from specified package
                if item in class_to_methods:
                    methods_to_run[item] = class_to_methods[item]
                continue

            # Check match for the class together with method specified
            class_to_test, test_method = item.rsplit('.', 1)
            if class_to_test in package.classes:
                methods_to_run[class_to_test] = [
                    m for m in class_to_methods[class_to_test]
                    if m == test_method]
                continue
        methods_count = sum(len(v) for v in methods_to_run.values())
        methods = [k + '.' + method
                   for k, v in methods_to_run.items() for method in v]
        LOG.debug('{0} method(s) is(are) going to be executed: '
                  '\n{1}'.format(methods_count, '\n'.join(methods)))
        return methods_to_run

    def _get_test_cases_by_classes(self, package):
        """Build valid test cases list for each class in the provided package.

           Check, if test class and test case name are valid.
           Return class mappings to test cases.
        """
        class_to_methods = {}
        for pkg_class_name in package.classes:
            class_obj = package.find_class(pkg_class_name, False)
            base_class = package.find_class(BASE_CLASS)
            if not base_class.is_compatible(class_obj):
                LOG.debug('Class {0} is not inherited from {1}. '
                          'Skipping it.'.format(pkg_class_name, BASE_CLASS))
                continue

            # Exclude methods, that are not test cases.
            tests = []
            valid_test_name = TEST_CASE_NAME

            for m in class_obj.methods:
                if valid_test_name.match(m):
                    tests.append(m)
            class_to_methods[pkg_class_name] = tests
        return class_to_methods

    def _call_service_method(self, name, exc, obj):
        if name in obj.type.all_method_names:
            method = obj.type.find_single_method(name)
            method.scope = dsl_types.MethodScopes.Public
            LOG.debug('Executing: {0}.{1}'.format(obj.type.name, name))
            exc.run(obj.type, name, obj, (), {})

    def _validate_keystone_opts(self, args):
        ks_opts_to_config = {
            'auth_url': 'www_authenticate_uri',
            'username': 'admin_user',
            'password': 'admin_password',
            'project_name': 'admin_project_name'}

        ks_opts = {'auth_url': getattr(args, 'os_auth_url', None),
                   'username': getattr(args, 'os_username', None),
                   'password': getattr(args, 'os_password', None),
                   'project_name': getattr(args, 'os_project_name', None)}

        if None in ks_opts.values() and not CONF.default_config_files:
            msg = ('Please provide murano config file or credentials for '
                   'authorization: {0}').format(
                ', '.join(['--os-auth-url', '--os-username', '--os-password',
                           '--os-project-name', '--os-tenant-id']))
            LOG.error(msg)
            self.error(msg)

        for param, value in ks_opts.items():
            if not value:
                ks_opts[param] = getattr(CONF.murano_auth,
                                         ks_opts_to_config[param])
            if param == 'auth_url':
                ks_opts[param] = ks_opts[param].replace('v2.0', 'v3')
        return ks_opts

    def error(self, msg, show_help=True):
        sys.stderr.write("ERROR: {msg}\n\n".format(msg=msg))
        if show_help:
            self.parser.print_help()
        sys.exit(1)

    def message(self, msg):
        sys.stdout.write('{0}\n'.format(msg))

    def run_tests(self):
        exit_code = 0
        provided_pkg_name = self.args.package
        load_packages_from = self.args.load_packages_from
        tests_to_run = self.args.tests

        ks_opts = self._validate_keystone_opts(self.args)

        client = ks_client.Client(**ks_opts)
        test_session = execution_session.ExecutionSession()
        test_session.token = client.auth_token
        test_session.project_id = client.project_id

        # Replace location of loading packages with provided from command line.
        if load_packages_from:
            cfg.CONF.engine.load_packages_from = load_packages_from
        with package_loader.CombinedPackageLoader(test_session) as pkg_loader:
            engine.get_plugin_loader().register_in_loader(pkg_loader)

            package = self._load_package(pkg_loader, provided_pkg_name)
            class_to_methods = self._get_test_cases_by_classes(package)
            run_set = self._get_methods_to_run(package,
                                               tests_to_run,
                                               class_to_methods)
            max_length = 0
            num_tests = 0
            for pkg_class, test_cases in run_set.items():
                for m in test_cases:
                    max_length = max(max_length, len(pkg_class) + len(m) + 1)
                num_tests += len(test_cases)
            max_length += 3

            if run_set:
                LOG.debug('Starting test execution.')
                self.message('About to execute {0} tests(s)'.format(num_tests))
            else:
                msg = _('No tests found for execution.')
                LOG.error(msg)
                self.error(msg)

            run_count = 0
            error_count = 0
            started = timeutils.utcnow()
            for pkg_class, test_cases in run_set.items():
                for m in test_cases:
                    # Create new executor for each test case to provide
                    # pure test environment
                    dsl_executor = executor.MuranoDslExecutor(
                        pkg_loader,
                        mock_context_manager.MockContextManager(),
                        test_session)
                    obj = dsl_executor.object_store.load(
                        {}, None,
                        default_type=package.find_class(pkg_class, False))

                    test_name = "{0}.{1}".format(obj.type.name, m)
                    dots_number = max_length - len(test_name)
                    msg = "{0} {1} ".format(test_name, '.' * dots_number)
                    sys.stdout.write(msg)
                    sys.stdout.flush()
                    self._call_service_method('setUp', dsl_executor, obj)
                    obj.type.methods[m].usage = 'Action'
                    test_session.start()
                    try:
                        run_count += 1
                        dsl_executor.run(obj.type, m, obj, (), {})
                        self._call_service_method(
                            'tearDown', dsl_executor, obj)
                        msg = '{0}{1}{2}\n'.format(OK_COLOR, 'OK', END_COLOR)
                        LOG.debug('Test {0} successful'.format(test_name))
                        sys.stdout.write(msg)
                        sys.stdout.flush()
                    except Exception as e:
                        error_count += 1
                        msg = ''.join((
                            FAIL_COLOR, 'FAIL!', END_COLOR, '\n'))
                        sys.stdout.write(msg)
                        if isinstance(e, dsl_exception.MuranoPlException):
                            tb = e.format()
                        else:
                            tb = traceback.format_exc()

                        sys.stdout.write(''.join((
                            FAIL_COLOR,
                            tb,
                            END_COLOR,
                            '\n'
                        )))
                        sys.stdout.flush()

                        LOG.exception('Test {0} failed'.format(test_name))
                        exit_code = 1
                    finally:
                        test_session.finish()
            completed = timeutils.utcnow()
            self.message('Executed {0} tests in {1} seconds: '
                         '{2} passed, '
                         '{3} failed'.format(run_count,
                                             timeutils.delta_seconds(
                                                 started, completed),
                                             run_count - error_count,
                                             error_count))
        return exit_code

    def get_parser(self):
        parser = argparse.ArgumentParser(prog='murano-test-runner')
        parser.set_defaults(func=self.run_tests)
        parser.add_argument('--config-file',
                            help='Path to the murano config')

        parser.add_argument('--os-auth-url',
                            default=utils.env('OS_AUTH_URL'),
                            help='Defaults to env[OS_AUTH_URL]')
        parser.add_argument('--os-username',
                            default=utils.env('OS_USERNAME'),
                            help='Defaults to env[OS_USERNAME]')

        parser.add_argument('--os-password',
                            default=utils.env('OS_PASSWORD'),
                            help='Defaults to env[OS_PASSWORD]')

        parser.add_argument('--os-project-name',
                            default=utils.env('OS_PROJECT_NAME'),
                            help='Defaults to env[OS_PROJECT_NAME]')
        parser.add_argument('-l', '--load_packages_from',
                            nargs='*', metavar='</path1, /path2>',
                            help='Directory to search packages from. '
                                 'We be added to the list of current directory'
                                 ' list, provided in a config file.')
        parser.add_argument("-v", "--verbose", action="store_true",
                            help="increase output verbosity")
        parser.add_argument('--version', action='version',
                            version=version.version_string)
        parser.add_argument('package',
                            metavar='<PACKAGE_FQN>',
                            help='Full name of application package that is '
                                 'going to be tested')
        parser.add_argument('tests', nargs='*',
                            metavar='<testMethod1, className.testMethod2>',
                            help='List of method names to be tested')
        return parser


def main():
    test_runner = MuranoTestRunner()
    try:
        if test_runner.args.config_file:
            default_config_files = [test_runner.args.config_file]
        else:
            default_config_files = cfg.find_config_files('murano')
            if not default_config_files:
                murano_conf = os.path.join(os.path.dirname(__file__),
                                           os.pardir,
                                           os.pardir,
                                           'etc', 'murano',
                                           'murano.conf')
                if os.path.exists(murano_conf):
                    default_config_files = [murano_conf]
        sys.argv = [sys.argv[0]]
        config.parse_args(default_config_files=default_config_files)
        CONF.set_default('use_stderr', False)
        logging.setup(CONF, 'murano')
    except RuntimeError as e:
        LOG.exception("Failed to initialize murano-test-runner: %s", e)
        sys.exit("ERROR: %s" % e)

    try:
        exit_code = test_runner.run_tests()
        sys.exit(exit_code)
    except Exception as e:
        if isinstance(e, dsl_exception.MuranoPlException):
            tb = e.format()
        else:
            tb = traceback.format_exc()
        err_msg = "Command failed: {0}\n{1}".format(e, tb)
        LOG.error(err_msg)
        sys.exit(err_msg)


if __name__ == '__main__':
    main()



# Copyright (c) 2015 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from murano.tests.unit.dsl.foundation import object_model as om
from murano.tests.unit.dsl.foundation import test_case


class TestCall(test_case.DslTestCase):
    def setUp(self):
        super(TestCall, self).setUp()
        self._runner = self.new_runner(om.Object('TestCall'))

    def test_call(self):
        self._runner.testCall()
        self.assertEqual(['called as call'], self._traces)

    def test_method(self):
        self._runner.testMethodInvocation()
        self.assertEqual(['called as method'], self._traces)

    def test_call_static(self):
        self._runner.testCallStatic()
        self.assertEqual(['called as static'], self._traces)

    def test_call_static_as_instance(self):
        self._runner.testCallStaticAsInstance()
        self.assertEqual(['called as static'], self._traces)



# Copyright (c) 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from yaql.language import exceptions
from yaql.language import specs
from yaql.language import yaqltypes

from murano.dsl import dsl
from murano.tests.unit.dsl.foundation import object_model as om
from murano.tests.unit.dsl.foundation import test_case


class TestExtensionMethods(test_case.DslTestCase):
    def setUp(self):
        @dsl.name('extcls.Extender')
        class PythonClass(object):
            def __init__(self, arg):
                self.value = arg

            @staticmethod
            @specs.meta('Usage', 'Extension')
            @specs.parameter('arg', yaqltypes.Integer())
            def python_extension(arg):
                return arg * arg

            @classmethod
            @specs.meta('Usage', 'Extension')
            @specs.parameter('arg', yaqltypes.Integer())
            def python_extension2(cls, arg):
                return cls(2 * arg).value

        super(TestExtensionMethods, self).setUp()
        self.package_loader.load_class_package(
            'extcls.Extender', None).register_class(PythonClass)

        self._runner = self.new_runner(om.Object('extcls.TestClass'))

    def test_call_self_extension_method(self):
        self.assertEqual([123, 123], self._runner.testSelfExtensionMethod())

    def test_call_imported_extension_method(self):
        self.assertEqual(
            [246, 246], self._runner.testImportedExtensionMethod())

    def test_call_nullable_extension_method(self):
        self.assertEqual(
            [123, None], self._runner.testNullableExtensionMethod())

    def test_extensions_precedence(self):
        self.assertEqual(111, self._runner.testExtensionsPrecedence())

    def test_explicit_call(self):
        self.assertEqual(222, self._runner.testCallExtensionExplicitly())

    def test_explicit_call_on_instance_fails(self):
        self.assertRaises(
            exceptions.NoMatchingMethodException,
            self._runner.testExplicitCallDoenstWorkOnInstance)

    def test_call_on_primitive_types(self):
        self.assertEqual('qWERTy', self._runner.testCallOnPrimitiveTypes())

    def test_call_python_extension(self):
        self.assertEqual(16, self._runner.testCallPythonExtension())

    def test_call_python_extension_explicitly(self):
        self.assertEqual(25, self._runner.testCallPythonExtensionExplicitly())

    def test_call_python_classmethod_extension(self):
        self.assertEqual(14, self._runner.testCallPythonClassmethodExtension())



#    Copyright (c) 2015 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json
import uuid

from oslo_config import cfg
from oslo_log import log as logging
import tenacity
from webob import response

from murano.common import auth_utils  # noqa
from murano.common import wsgi
from murano.db.services import cf_connections as db_cf
import muranoclient.client as muranoclient
from muranoclient.common import exceptions
from muranoclient.glance import client as glare_client


LOG = logging.getLogger(__name__)
CONF = cfg.CONF


class Controller(object):
    """WSGI controller for application catalog resource in Murano v1 API"""

    def _package_to_service(self, package):
        srv = {}
        srv['id'] = package.id
        srv['name'] = package.name
        if len(package.description) > 256:
            srv['description'] = u"{0} ...".format(package.description[:253])
        else:
            srv['description'] = package.description
        srv['bindable'] = True
        srv['tags'] = []
        for tag in package.tags:
            srv['tags'].append(tag)
        plan = {'id': package.id + '-1',
                'name': 'default',
                'description': 'Default plan for the service {name}'.format(
                    name=package.name)}
        srv['plans'] = [plan]
        return srv

    def _make_service(self, name, package, plan_id):
        id = uuid.uuid4().hex

        return {"name": name,
                "?": {plan_id: {"name": package.name},
                      "type": package.fully_qualified_name,
                      "id": id}}

    def _get_service(self, env, service_id):
        for service in env.services:
            if service['?']['id'] == service_id:
                return service
        return None

    def list(self, req):

        token = req.headers['X-Auth-Token']
        m_cli = _get_muranoclient(token, req)
        kwargs = {'type': 'Application'}
        packages = m_cli.packages.filter(**kwargs)
        services = []
        for package in packages:
            services.append(self._package_to_service(package))

        resp = {'services': services}

        return resp

    def provision(self, req, body, instance_id):
        """Here is the example of request body given us from Cloud Foundry:

         {
         "service_id":        "service-guid-here",
         "plan_id":           "plan-guid-here",
         "organization_guid": "org-guid-here",
         "space_guid":        "space-guid-here",
         "parameters": {"param1": "value1",
                        "param2": "value2"}
         }
        """
        data = json.loads(req.body)
        space_guid = data['space_guid']
        org_guid = data['organization_guid']
        plan_id = data['plan_id']
        service_id = data['service_id']
        parameters = data['parameters']
        self.current_session = None

        # Here we'll take an entry for CF org and space from db. If we
        # don't have any entries we will create it from scratch.
        try:
            tenant = db_cf.get_tenant_for_org(org_guid)
        except AttributeError:
            tenant = req.headers['X-Project-Id']
            db_cf.set_tenant_for_org(org_guid, tenant)
            LOG.info("Cloud Foundry {org_id} mapped to tenant "
                     "{tenant_name}".format(org_id=org_guid,
                                            tenant_name=tenant))

        token = req.headers['X-Auth-Token']
        m_cli = _get_muranoclient(token, req)

        def _set_new_environment_for_space(space_guid, log_msg):
            body = {'name': 'my_{uuid}'.format(uuid=uuid.uuid4().hex)}
            env = m_cli.environments.create(body)
            db_cf.set_environment_for_space(space_guid, env.id)
            LOG.info(log_msg.format(space_id=space_guid,
                                    environment_id=env.id))
            return env.id

        try:
            environment_id = db_cf.get_environment_for_space(space_guid)
            # NOTE: Check that environment which was previously linked with
            # CF space still exist, reset a new environment for space.
            try:
                env = m_cli.environments.get(environment_id)
            except exceptions.HTTPNotFound:
                msg = ("Can not find environment_id {environment_id}, "
                       "will create a new one."
                       .format(environment_id=environment_id))
                LOG.info(msg)
                env = {}
            if not env:
                log_msg = ("Cloud Foundry {space_id} remapped to "
                           "{environment_id}")
                environment_id = _set_new_environment_for_space(
                    space_guid, log_msg)
        except AttributeError:
            log_msg = ("Cloud Foundry {space_id} mapped to "
                       "{environment_id}")
            environment_id = _set_new_environment_for_space(
                space_guid, log_msg)

        package = m_cli.packages.get(service_id)
        LOG.debug('Adding service {name}'.format(name=package.name))

        service = self._make_service(space_guid, package, plan_id)
        db_cf.set_instance_for_service(instance_id, service['?']['id'],
                                       environment_id, tenant)

        # NOTE(Kezar): Here we are going through JSON and add ids where
        # it's necessary. Before that we need to drop '?' key from parameters
        # dictionary as far it contains murano package related info which is
        # necessary in our scenario
        if '?' in parameters.keys():
            parameters.pop('?', None)
            LOG.warning("Incorrect input parameters. Package related "
                        "parameters shouldn't be passed through Cloud "
                        "Foundry")
        params = [parameters]
        while params:
            a = params.pop()
            for k, v in a.items():
                if isinstance(v, dict):
                    params.append(v)
                    if k == '?':
                        v['id'] = uuid.uuid4().hex
        service.update(parameters)
        # Now we need to obtain session to modify the env
        session_id = create_session(m_cli, environment_id)
        m_cli.services.post(environment_id,
                            path='/',
                            data=service,
                            session_id=session_id)
        m_cli.sessions.deploy(environment_id, session_id)
        self.current_session = session_id
        return response.Response(status=202, json_body={})

    def deprovision(self, req, instance_id):
        service = db_cf.get_service_for_instance(instance_id)
        if not service:
            return {}

        service_id = service.service_id
        environment_id = service.environment_id
        token = req.headers['X-Auth-Token']
        m_cli = _get_muranoclient(token, req)

        session_id = create_session(m_cli, environment_id)
        m_cli.services.delete(environment_id, '/' + service_id, session_id)
        m_cli.sessions.deploy(environment_id, session_id)
        return response.Response(status=202, json_body={})

    def bind(self, req, body, instance_id, app_id):
        db_service = db_cf.get_service_for_instance(instance_id)
        if not db_service:
            return {}

        service_id = db_service.service_id
        environment_id = db_service.environment_id
        token = req.headers['X-Auth-Token']
        m_cli = _get_muranoclient(token, req)

        session_id = create_session(m_cli, environment_id)
        env = m_cli.environments.get(environment_id, session_id)
        LOG.debug('Got environment {0}'.format(env))
        service = self._get_service(env, service_id)
        LOG.debug('Got service {0}'.format(service))

        # NOTE(starodubcevna): Here we need to find an action which will return
        # us needed credentials. By default we will look for getCredentials
        # action.
        result = {}
        try:
            actions = service['?']['_actions']
            for action_id in list(actions):
                if 'getCredentials' in action_id:

                    @tenacity.retry(
                        retry=tenacity.retry_if_exception_type(TypeError),
                        wait=tenacity.wait_random(min=1, max=10),
                        stop=tenacity.stop_after_delay(30),
                        reraise=True)
                    def _get_creds(client, task_id, environment_id):
                        result = m_cli.actions.get_result(environment_id,
                                                          task_id)['result']
                        return result

                    task_id = m_cli.actions.call(environment_id, action_id)
                    result = _get_creds(m_cli, task_id, environment_id)

            if not result:
                LOG.warning("This application doesn't have action "
                            "getCredentials")
                return response.Response(status=500)
        except KeyError:
            # NOTE(starodubcevna): In CF service broker API spec return
            # code for failed bind is not present, so we will return 500.
            LOG.warning("This application doesn't have actions at all")
            return response.Response(status=500)

        if 'credentials' in list(result):
            return result
        else:
            return {'credentials': result}

    def unbind(self, req, instance_id, app_id):
        """Unsupported functionality

        murano doesn't support this kind of functionality, so we just need
        to create a stub where the call will come. We can't raise something
        like NotImplementedError because we will have problems on Cloud Foundry
        side. The best way now it to return empty dict which will be correct
        answer for Cloud Foundry.
        """

        return {}

    def get_last_operation(self, req, instance_id):
        service = db_cf.get_service_for_instance(instance_id)
        # NOTE(freerunner): Prevent code 500 if requested environment
        # already doesn't exist.
        if not service:
            LOG.warning('Requested service for instance {0} is not '
                        'found'.format(instance_id))
            body = {}
            resp = response.Response(status=410, json_body=body)
            return resp
        env_id = service.environment_id
        token = req.headers["X-Auth-Token"]
        m_cli = _get_muranoclient(token, req)

        # NOTE(starodubcevna): we can track only environment status. it's
        # murano API limitation.
        m_environment = m_cli.environments.get(env_id)
        body = {'state': 'unknown', 'description': 'operation unknown'}
        resp = response.Response(status=500, json_body=body)
        if m_environment.status == 'ready':
            body = {'state': 'succeeded',
                    'description': 'operation succeed'}
            resp = response.Response(status=200, json_body=body)
        elif m_environment.status in ['pending', 'deleting', 'deploying']:
            body = {'state': 'in progress',
                    'description': 'operation in progress'}
            resp = response.Response(status=202, json_body=body)
        elif m_environment.status in ['deploy failure', 'delete failure']:
            body = {'state': 'failed',
                    'description': '{0}. Please correct it manually'.format(
                        m_environment.status)}
            resp = response.Response(status=200, json_body=body)
        return resp


def _get_muranoclient(token_id, req):

    artifacts_client = None
    if CONF.engine.packages_service in ['glance', 'glare']:
        artifacts_client = _get_glareclient(token_id, req)

    murano_url = CONF.murano.url or req.endpoints.get('murano')
    if not murano_url:
        LOG.error('No murano url is specified and no '
                  '"application-catalog" '
                  'service is registered in keystone.')

    return muranoclient.Client(1, murano_url, token=token_id,
                               artifacts_client=artifacts_client)


def _get_glareclient(token_id, req):
    glare_settings = CONF.glare

    url = glare_settings.url or req.endpoints.get('glare')
    if not url:
        LOG.error('No glare url is specified and no "artifact" '
                  'service is registered in keystone.')
    # TODO(gyurco): use auth_utils.get_session_client_parameters
    return glare_client.Client(
        endpoint=url, token=token_id,
        insecure=glare_settings.insecure,
        key_file=glare_settings.keyfile or None,
        ca_file=glare_settings.cafile or None,
        cert_file=glare_settings.certfile or None,
        type_name='murano',
        type_version=1)


def create_session(client, environment_id):
    id = client.sessions.configure(environment_id).id
    return id


def create_resource():
    return wsgi.Resource(Controller(),
                         serializer=wsgi.ServiceBrokerResponseSerializer())



#    Copyright (c) 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock
import weakref

from testtools import matchers

from murano.dsl import exceptions
from murano.dsl import murano_method
from murano.dsl import murano_object
from murano.dsl import murano_type
from murano.dsl.principal_objects import garbage_collector
from murano.tests.unit import base


class TestGarbageCollector(base.MuranoTestCase):
    def setUp(self):
        super(TestGarbageCollector, self).setUp()

        self.subscriber = mock.MagicMock(spec=murano_object.MuranoObject)
        self.subscriber.real_this = self.subscriber

        mock_class = mock.MagicMock(spec=murano_type.MuranoClass)
        mock_method = mock.MagicMock(spec=murano_method.MuranoMethod)
        mock_method.name = "mockHandler"
        mock_class.methods = mock.PropertyMock(
            return_value={"mockHandler": mock_method})

        def find_single_method(name):
            if name != 'mockHandler':
                raise exceptions.NoMethodFound(name)

        mock_class.find_single_method = find_single_method
        self.subscriber.type = mock_class

        self.publisher = mock.MagicMock(spec=murano_object.MuranoObject)
        self.publisher.real_this = self.publisher

    def test_set_dd(self):
        self.publisher.destruction_dependencies = []
        garbage_collector.GarbageCollector.subscribe_destruction(
            self.publisher, self.subscriber, handler="mockHandler")
        dep = self.publisher.destruction_dependencies
        self.assertThat(dep, matchers.HasLength(1))
        dep = dep[0]
        self.assertEqual("mockHandler", dep["handler"])
        self.assertEqual(self.subscriber, dep["subscriber"]())

    def test_unset_dd(self):
        self.publisher.destruction_dependencies = [{
            "subscriber": weakref.ref(self.subscriber),
            "handler": "mockHandler"
        }]
        garbage_collector.GarbageCollector.unsubscribe_destruction(
            self.publisher, self.subscriber, handler="mockHandler")
        self.assertEqual(
            [], self.publisher.destruction_dependencies)

    def test_set_wrong_handler(self):
        self.assertRaises(
            exceptions.NoMethodFound,
            garbage_collector.GarbageCollector.subscribe_destruction,
            self.publisher, self.subscriber, handler="invalidHandler")
        self.assertRaises(
            exceptions.NoMethodFound,
            garbage_collector.GarbageCollector.unsubscribe_destruction,
            self.publisher, self.subscriber, handler="invalidHandler")



#    Copyright (c) 2013 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from murano.common.helpers import token_sanitizer
from murano.tests.unit import base


class TokenSanitizerTests(base.MuranoTestCase):
    sanitizer = token_sanitizer.TokenSanitizer()

    def test_dict_with_one_value(self):
        source = {'token': 'value'}
        value = self.sanitizer.sanitize(source)
        self.assertEqual(self.sanitizer.message, value['token'])

    def test_dict_with_few_value(self):
        source = {'token': 'value', 'pass': 'value', 'TrustId': 'value'}
        value = self.sanitizer.sanitize(source)

        self.assertEqual(self.sanitizer.message, value['token'])
        self.assertEqual(self.sanitizer.message, value['pass'])
        self.assertEqual(self.sanitizer.message, value['TrustId'])

    def test_dict_with_nested_dict(self):
        source = {'obj': {'pass': 'value'}}
        value = self.sanitizer.sanitize(source)
        self.assertEqual(self.sanitizer.message, value['obj']['pass'])

    def test_dict_with_nested_list(self):
        source = {'obj': [{'pass': 'value'}]}
        value = self.sanitizer.sanitize(source)
        self.assertEqual(self.sanitizer.message, value['obj'][0]['pass'])

    def test_leave_out_other_values(self):
        source = {'obj': ['value']}
        value = self.sanitizer.sanitize(source)
        self.assertEqual('value', value['obj'][0])



#    Copyright (c) 2013 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

stats = None

SUPPORTED_PARAMS = {'id', 'order_by', 'category', 'marker', 'tag',
                    'class_name', 'limit', 'type', 'fqn', 'category', 'owned',
                    'search', 'include_disabled', 'sort_dir', 'name'}
LIST_PARAMS = {'id', 'category', 'tag', 'class', 'order_by'}
ORDER_VALUES = {'fqn', 'name', 'created'}
OPERATOR_VALUES = {'id', 'category', 'tag'}
PKG_PARAMS_MAP = {'display_name': 'name',
                  'full_name': 'fully_qualified_name',
                  'ui': 'ui_definition',
                  'logo': 'logo',
                  'package_type': 'type',
                  'description': 'description',
                  'author': 'author',
                  'classes': 'class_definitions',
                  'tags': 'tags',
                  'supplier': 'supplier',
                  'supplier_logo': 'supplier_logo'}



# Copyright (c) 2016 AT&T Corp
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import random
import shutil
import string
import tempfile
from unittest import mock
import yaml
import zipfile

from murano.packages import exceptions
from murano.packages import load_utils
import murano.tests.unit.base as test_base


class TestLoadUtils(test_base.MuranoTestCase):

    def setUp(cls):
        super(TestLoadUtils, cls).setUp()
        cls.temp_directories = []
        cls.temp_files = []

    def _create_temp_dir(self):
        temp_directory = tempfile.mkdtemp()
        self.temp_directories.append(temp_directory)
        return temp_directory

    def _create_temp_file(self):
        temp_file = tempfile.NamedTemporaryFile(delete=True)
        self.temp_files.append(temp_file)
        return temp_file

    def _create_temp_zip_file(self, zip_path, manifest_path,
                              arcname='manifest.yaml'):
        zip_ = zipfile.ZipFile(zip_path, 'w')
        zip_.write(manifest_path, arcname=arcname)
        zip_.close()
        self.temp_files.append(zip_)
        return zip_

    def tearDown(cls):
        super(TestLoadUtils, cls).tearDown()
        for directory in cls.temp_directories:
            if os.path.isdir(directory):
                shutil.rmtree(directory)
        for file in cls.temp_files:
            if isinstance(file, zipfile.ZipFile):
                if zipfile.is_zipfile(file.filename):
                    os.remove(file)
            else:
                if os.path.isfile(file.name):
                    os.remove(file.name)

    def _test_load_from_file(self, target_dir=None, drop_dir=True):
        manifest_file_contents = dict(
            Format='MuranoPL/1.1',
            FullName='test_full_name',
            Type='Application',
            Description='test_description',
            Author='test_author',
            Supplier='test_supplier',
            Tags=[]
        )
        test_directory = self._create_temp_dir()
        manifest_path = os.path.join(test_directory, 'manifest.yaml')
        zip_path = os.path.join(test_directory, 'test_zip_load_utils.zip')

        with open(manifest_path, 'w') as manifest_file:
            yaml.dump(manifest_file_contents, manifest_file,
                      default_flow_style=True)
            self._create_temp_zip_file(zip_path, manifest_path)

        with load_utils.load_from_file(archive_path=zip_path,
                                       target_dir=target_dir,
                                       drop_dir=drop_dir) as plugin:
            self.assertEqual('MuranoPL', plugin.format_name)
            self.assertEqual('1.1.0', str(plugin.runtime_version))
            self.assertEqual(manifest_file_contents['FullName'],
                             plugin.full_name)
            self.assertEqual(manifest_file_contents['Description'],
                             plugin.description)
            self.assertEqual(manifest_file_contents['Author'],
                             plugin.author)
            self.assertEqual(manifest_file_contents['Supplier'],
                             plugin.supplier)
            self.assertEqual(manifest_file_contents['Tags'],
                             plugin.tags)

    def test_load_from_file(self):
        self._test_load_from_file(target_dir=None, drop_dir=True)

    def test_load_from_file_with_custom_target_directory(self):
        target_dir = self._create_temp_dir()
        self._test_load_from_file(target_dir=target_dir, drop_dir=True)

    @mock.patch('murano.packages.load_utils.get_plugin_loader')
    def test_load_from_file_with_invalid_handler(self, mock_plugin_loader):
        mock_plugin_loader().get_package_handler = mock.MagicMock(
            return_value=None)
        test_format = 'Invalid Format'
        manifest_file_contents = dict(
            Format=test_format,
            FullName='test_full_name',
            Type='Application',
            Description='test_description',
            Author='test_author',
            Supplier='test_supplier',
            Tags=[]
        )
        test_directory = self._create_temp_dir()
        target_dir = self._create_temp_dir()
        manifest_path = os.path.join(test_directory, 'manifest.yaml')
        zip_path = os.path.join(test_directory, 'test_zip_load_utils.zip')

        with open(manifest_path, 'w') as manifest_file:
            yaml.dump(manifest_file_contents, manifest_file,
                      default_flow_style=True)
            self._create_temp_zip_file(zip_path, manifest_path)

        expected_error_msg = "Unsupported format {0}".format(test_format)
        with self.assertRaisesRegex(exceptions.PackageLoadError,
                                    expected_error_msg):
            with load_utils.load_from_file(archive_path=zip_path,
                                           target_dir=target_dir,
                                           drop_dir=True):
                pass
        mock_plugin_loader().get_package_handler.assert_called_once_with(
            test_format)

    def test_load_from_file_with_invalid_archive_path(self):
        expected_error_msg = "Unable to find package file"
        with self.assertRaisesRegex(exceptions.PackageLoadError,
                                    expected_error_msg):
            with load_utils.load_from_file('invalid file path'):
                pass

    @mock.patch('murano.packages.load_utils.os')
    def test_load_from_file_with_nonempty_target_directory(self, mock_os):
        mock_os.listdir = mock.MagicMock(return_value=True)
        temp_file = self._create_temp_file()
        expected_error_msg = "Target directory is not empty"
        with self.assertRaisesRegex(exceptions.PackageLoadError,
                                    expected_error_msg):
            this_dir = os.path.dirname(os.path.realpath(__file__))
            with load_utils.load_from_file(temp_file.name,
                                           target_dir=this_dir):
                pass

    def test_load_from_file_without_zip_file(self):
        temp_file = self._create_temp_file()
        expected_error_msg = "Uploaded file {0} is not a zip archive".\
                             format(temp_file.name)
        with self.assertRaisesRegex(exceptions.PackageLoadError,
                                    expected_error_msg):
            with load_utils.load_from_file(temp_file.name):
                pass

    @mock.patch('murano.packages.load_utils.zipfile')
    def test_load_from_file_handle_value_error(self, mock_zipfile):
        test_error_msg = 'Random error message.'
        expected_error_msg = "Couldn't load package from file: {0}".\
                             format(test_error_msg)
        mock_zipfile.is_zipfile = mock.MagicMock(
            side_effect=ValueError(test_error_msg))
        temp_file = self._create_temp_file()
        with self.assertRaisesRegex(exceptions.PackageLoadError,
                                    expected_error_msg):
            with load_utils.load_from_file(temp_file.name):
                pass
        mock_zipfile.is_zipfile.assert_called_once_with(
            temp_file.name)

    def test_load_from_dir_without_source_directory(self):
        expected_error_msg = 'Invalid package directory'
        with self.assertRaisesRegex(exceptions.PackageLoadError,
                                    expected_error_msg):
            load_utils.load_from_dir('random_test_directory')

    def test_load_from_dir_with_invalid_source_directory(self):
        source_directory = self._create_temp_dir()
        expected_error_msg = 'Unable to find package manifest'
        with self.assertRaisesRegex(exceptions.PackageLoadError,
                                    expected_error_msg):
            load_utils.load_from_dir(source_directory)

    @mock.patch('murano.packages.load_utils.os.path.isfile')
    def test_load_from_dir_open_file_negative(self, mock_isfile):
        mock_isfile.return_value = True
        source_directory = self._create_temp_dir()
        random_filename = ''.join(random.choice(string.ascii_lowercase)
                                  for i in range(20))
        expected_error_msg = 'Unable to load due to'
        with self.assertRaisesRegex(exceptions.PackageLoadError,
                                    expected_error_msg):
            load_utils.load_from_dir(source_directory,
                                     filename=random_filename)



# Copyright 2016 AT&T Corp
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections
from muranoclient.common import exceptions as muranoclient_exc
import os
import shutil
import tempfile
from unittest import mock

import semantic_version
import testtools

from murano.dsl import exceptions as dsl_exceptions
from murano.dsl import murano_package as dsl_package
from murano.engine import package_loader
from murano.packages import exceptions as pkg_exc
from murano.tests.unit import base
from murano.tests.unit import utils


class TestPackageCache(base.MuranoTestCase):

    def setUp(self):
        super(TestPackageCache, self).setUp()

        self.location = tempfile.mkdtemp()
        self.override_config('enable_packages_cache', True, 'engine')
        self.override_config('packages_cache', self.location, 'engine')

        self._patch_loader_client()
        self.loader = package_loader.ApiPackageLoader(None)

    def tearDown(self):
        shutil.rmtree(self.location, ignore_errors=True)
        super(TestPackageCache, self).tearDown()

    def _patch_loader_client(self):
        self.murano_client_patcher = mock.patch(
            'murano.engine.package_loader.ApiPackageLoader.client')
        self.murano_client_patcher.start()
        self.murano_client = package_loader.ApiPackageLoader.client

    def _unpatch_loader_client(self):
        self.murano_client_patcher.stop()

    @mock.patch('murano.engine.package_loader.auth_utils')
    @mock.patch('murano.engine.package_loader.versionutils')
    def test_client_property(self, mock_versionutils, mock_auth_utils):
        self._unpatch_loader_client()
        session = mock_auth_utils.get_client_session()
        session_params = mock_auth_utils.get_session_client_parameters
        session.auth.get_token.return_value = 'test_token'
        session.get_endpoint.return_value = 'test_endpoint/v3'
        session_params.return_value = {'endpoint': 'test_endpoint/v3'}

        self.override_config('packages_service', 'glance', group='engine')

        client = self.loader.client

        mock_versionutils.report_deprecated_feature.assert_called_once_with(
            package_loader.LOG,
            "'glance' packages_service option has been renamed "
            "to 'glare', please update your configuration")

        self.assertIsNotNone(client)
        self.assertIsNotNone(self.loader._glare_client)

        # Test whether client is initialized using different CONF.
        self.override_config('packages_service', 'test_service',
                             group='engine')

        client = self.loader.client

        self.assertIsNotNone(client)

    def test_import_fixations_table(self):
        test_fixations = {
            'test_package_1': [semantic_version.Version('1.1.0'),
                               semantic_version.Version('1.1.0')],
            'test_package_2': [semantic_version.Version('2.1.0'),
                               semantic_version.Version('2.4.3')]
        }
        expected = collections.defaultdict(set)
        expected['test_package_1'] = set([semantic_version.Version('1.1.0')])
        expected['test_package_2'] = set([semantic_version.Version('2.1.0'),
                                          semantic_version.Version('2.4.3')])

        self.loader.import_fixation_table(test_fixations)
        self.assertEqual(expected, self.loader._fixations)

    def test_register_package(self):
        test_version = semantic_version.Version('1.1.0')
        package = mock.MagicMock()
        package.name = 'test_package_name'
        package.version = test_version
        package.classes = ['test_class_1', 'test_class_2']

        self.loader.register_package(package)
        self.assertEqual(
            package,
            self.loader._package_cache['test_package_name'][test_version])
        for class_name in package.classes:
            self.assertEqual(
                package,
                self.loader._class_cache[class_name][test_version])

    def test_load_package(self):
        test_version = semantic_version.Version('1.1.0')

        package = mock.MagicMock()
        package.name = 'test_package_name'
        package.version = test_version
        self.loader.import_fixation_table({package.name: [test_version]})
        self.loader.register_package(package)

        version_spec = semantic_version.Spec('>=1.0.0,<2.4.0')
        retrieved_pkg = self.loader.load_package(package.name, version_spec)
        self.assertEqual(retrieved_pkg, package)

    @testtools.skipIf(os.name == 'nt', "Doesn't work on Windows")
    @mock.patch('murano.engine.package_loader.ApiPackageLoader.'
                '_to_dsl_package')
    def test_load_package_with_get_definiton(self, mock_to_dsl_package):
        fqn = 'io.murano.apps.test_package'
        package = mock.MagicMock()
        package.id = 'test_package_id'
        package.name = 'test_package_name'
        package.fully_qualified_name = fqn
        package.version = '2.5.3'

        path, _ = utils.compose_package(
            'test_package', self.location, archive_dir=self.location,
            version=package.version)
        with open(path, 'rb') as f:
            package_data = f.read()

        self.murano_client.packages.filter = mock.MagicMock(
            return_value=[package])
        self.murano_client.packages.download = mock.MagicMock(
            return_value=package_data)
        mock_to_dsl_package.return_value = package

        spec = semantic_version.Spec('*')
        retrieved_pkg = self.loader.load_package(fqn, spec)
        self.assertEqual(retrieved_pkg, package)

        self.assertTrue(os.path.isdir(os.path.join(
            self.location, fqn)))
        self.assertTrue(os.path.isdir(os.path.join(
            self.location, fqn, package.version)))
        self.assertTrue(os.path.isdir(os.path.join(
            self.location, fqn, package.version, package.id)))
        self.assertTrue(os.path.isfile(os.path.join(
            self.location, fqn, package.version, package.id, 'manifest.yaml')))
        self.murano_client.packages.download.assert_called_once_with(
            package.id)

        expected_fixations = collections.defaultdict(set)
        expected_fixations[fqn] = set(
            [semantic_version.Version(package.version)])
        self.assertEqual(expected_fixations, self.loader._fixations)
        self.loader.cleanup()

    def test_load_package_except_lookup_error(self):
        expected_error_msg = 'Package "test_package_name" is not found'
        invalid_specs = [semantic_version.Spec('>=1.1.1'),
                         semantic_version.Spec('<1.1.0'),
                         semantic_version.Spec('>=1.1.1,<1.1.0'),
                         semantic_version.Spec('==1.1.1')]

        fqn = 'io.murano.apps.test'
        test_version = semantic_version.Version('1.1.0')
        package = mock.MagicMock()
        package.name = fqn
        package.fully_qualified_name = fqn
        package.version = test_version
        self.loader.import_fixation_table({fqn: [test_version]})
        self.loader.register_package(package)

        with self.assertRaisesRegex(dsl_exceptions.NoPackageFound,
                                    expected_error_msg):
            for spec in invalid_specs:
                self.loader.load_package('test_package_name', spec)

    def test_load_class_package(self):
        fqn = 'io.murano.apps.test'
        package = mock.MagicMock()
        package.fully_qualified_name = fqn
        package.classes = ['test_class_1', 'test_class_2']
        test_version = semantic_version.Version('1.1.0')
        package.version = test_version
        self.loader.register_package(package)

        spec = semantic_version.Spec('*')
        for class_name in ['test_class_1', 'test_class_2']:
            retrieved_pkg = self.loader.load_class_package(class_name, spec)
            self.assertEqual(retrieved_pkg, package)

    @testtools.skipIf(os.name == 'nt', "Doesn't work on Windows")
    def test_load_class_package_with_get_definition(self):
        fqn = 'io.murano.apps.test'
        path, name = utils.compose_package(
            'test',
            self.location, archive_dir=self.location)
        with open(path, 'rb') as f:
            package_data = f.read()
        spec = semantic_version.Spec('*')

        first_id, second_id, third_id = '123', '456', '789'
        package = mock.MagicMock()
        package.fully_qualified_name = fqn
        package.id = first_id
        package.version = '0.0.1'

        self.murano_client.packages.filter = mock.MagicMock(
            return_value=[package])
        self.murano_client.packages.download = mock.MagicMock(
            return_value=package_data)

        # load the package
        self.loader.load_class_package(fqn, spec)

        # assert that everything got created
        self.assertTrue(os.path.isdir(os.path.join(
            self.location, fqn)))
        self.assertTrue(os.path.isdir(os.path.join(
            self.location, fqn, package.version)))
        self.assertTrue(os.path.isdir(os.path.join(
            self.location, fqn, package.version, first_id)))
        self.assertTrue(os.path.isfile(os.path.join(
            self.location, fqn, package.version, first_id, 'manifest.yaml')))

        # assert that we called download
        self.assertEqual(self.murano_client.packages.download.call_count, 1)

        # now that the cache is in place, call it for the 2d time
        self.loader._package_cache = {}
        self.loader._class_cache = {}
        self.loader.load_class_package(fqn, spec)

        # check that we didn't download a thing
        self.assertEqual(self.murano_client.packages.download.call_count, 1)

        # changing id, new package would be downloaded.
        package.id = second_id
        self.loader._package_cache = {}
        self.loader._class_cache = {}
        self.loader.load_class_package(fqn, spec)

        # check that we didn't download a thing
        self.assertEqual(self.murano_client.packages.download.call_count, 2)

        # check that old directories were not deleted
        # we did not call cleanup and did not release the locks
        self.assertTrue(os.path.isdir(os.path.join(
            self.location, fqn, package.version, first_id)))

        # check that new directories got created correctly
        self.assertTrue(os.path.isdir(os.path.join(
            self.location, fqn)))
        self.assertTrue(os.path.isdir(os.path.join(
            self.location, fqn, package.version)))
        self.assertTrue(os.path.isdir(os.path.join(
            self.location, fqn, package.version, second_id)))
        self.assertTrue(os.path.isfile(os.path.join(
            self.location, fqn, package.version, second_id, 'manifest.yaml')))

        self.assertTrue(os.path.isdir(os.path.join(
            self.location, fqn, package.version)))
        self.assertTrue(os.path.isdir(os.path.join(
            self.location, fqn, package.version, second_id)))

        # changing id, new package would be downloaded.
        package.id = third_id
        self.loader._package_cache = {}
        self.loader._class_cache = {}

        # release all the locks
        self.loader.cleanup()
        self.loader.load_class_package(fqn, spec)

        # check that we didn't download a thing
        self.assertEqual(self.murano_client.packages.download.call_count, 3)

        # check that old directories were *deleted*
        self.assertFalse(os.path.isdir(os.path.join(
            self.location, fqn, package.version, first_id)))
        self.assertFalse(os.path.isdir(os.path.join(
            self.location, fqn, package.version, second_id)))

        # check that new directories got created correctly
        self.assertTrue(os.path.isdir(os.path.join(
            self.location, fqn, package.version, third_id)))
        self.assertTrue(os.path.isfile(os.path.join(
            self.location, fqn, package.version, third_id, 'manifest.yaml')))

    def test_load_class_package_except_lookup_error(self):
        invalid_specs = [semantic_version.Spec('>=1.1.1'),
                         semantic_version.Spec('<1.1.0'),
                         semantic_version.Spec('>=1.1.1,<1.1.0'),
                         semantic_version.Spec('==1.1.1')]

        fqn = 'io.murano.apps.test'
        test_version = semantic_version.Version('1.1.0')
        package = mock.MagicMock()
        package.name = fqn
        package.fully_qualified_name = fqn
        package.version = test_version
        package.classes = ['test_class_1', 'test_class_2']
        self.loader.import_fixation_table({fqn: [test_version]})
        self.loader.register_package(package)

        for class_ in package.classes:
            expected_error_msg = 'Package for class "{0}" is not found'\
                                 .format(class_)
            with self.assertRaisesRegex(dsl_exceptions.NoPackageForClassFound,
                                        expected_error_msg):
                for spec in invalid_specs:
                    self.loader.load_class_package(class_, spec)

    @mock.patch('murano.engine.package_loader.LOG')
    def test_get_definition_with_multiple_packages_returned(self, mock_log):
        self.loader._execution_session = mock.MagicMock()
        self.loader._execution_session.project_id = 'test_project_id'

        matching_pkg = mock.MagicMock(owner_id='test_project_id',
                                      is_public=False)
        public_pkg = mock.MagicMock(owner_id='another_test_project_id',
                                    is_public=True)
        other_pkg = mock.MagicMock(owner_id='another_test_project_id',
                                   is_public=False)

        # Test package with matching owner_id is returned, despite ordering.
        self.loader.client.packages.filter.return_value =\
            [public_pkg, matching_pkg, other_pkg]
        expected_pkg = matching_pkg
        retrieved_pkg = self.loader._get_definition({})
        self.assertEqual(expected_pkg, retrieved_pkg)

        # Test public package is returned, despite ordering.
        self.loader.client.packages.filter.return_value =\
            [other_pkg, public_pkg]
        expected_pkg = public_pkg
        retrieved_pkg = self.loader._get_definition({})
        self.assertEqual(expected_pkg, retrieved_pkg)

        # Test other package is returned.
        self.loader.client.packages.filter.return_value =\
            [other_pkg, other_pkg]
        expected_pkg = other_pkg
        retrieved_pkg = self.loader._get_definition({})
        self.assertEqual(expected_pkg, retrieved_pkg)

        mock_log.debug.assert_any_call(
            'Ambiguous package resolution: more than 1 package found for query'
            ' "{opts}", will resolve based on the ownership'
            .format(opts={'catalog': True}))

    @mock.patch('murano.engine.package_loader.LOG')
    def test_get_definition_except_lookup_error(self, mock_log):
        self.loader.client.packages.filter.return_value = []

        with self.assertRaisesRegex(LookupError, None):
            self.loader._get_definition({})

        mock_log.debug.assert_called_once_with(
            "There are no packages matching filter {opts}"
            .format(opts={'catalog': True}))
        mock_log.debug.reset_mock()

        self.loader.client.packages.filter.side_effect =\
            muranoclient_exc.HTTPException

        with self.assertRaisesRegex(LookupError, None):
            self.loader._get_definition({})

        mock_log.debug.assert_called_once_with(
            'Failed to get package definition from repository')

    @testtools.skipIf(os.name == 'nt', "Doesn't work on Windows")
    @mock.patch('murano.engine.package_loader.LOG')
    @mock.patch('murano.engine.package_loader.os')
    @mock.patch('murano.engine.package_loader.load_utils')
    def test_get_package_by_definition_except_package_load_error(
            self, mock_load_utils, mock_os, mock_log):
        # Test that the first instance of the exception is caught.
        temp_directory = tempfile.mkdtemp(prefix='test-package-loader-',
                                          dir=tempfile.tempdir)
        mock_os.path.isdir.return_value = True
        mock_os.path.join.return_value = temp_directory
        mock_load_utils.load_from_dir.side_effect = pkg_exc.PackageLoadError

        fqn = 'io.murano.apps.test'
        path, _ = utils.compose_package(
            'test', self.location, archive_dir=self.location)
        with open(path, 'rb') as f:
            package_data = f.read()

        package = mock.MagicMock()
        package.fully_qualified_name = fqn
        package.id = '123'
        package.version = '0.0.1'

        self.murano_client.packages.download = mock.MagicMock(
            return_value=package_data)

        self.loader._get_package_by_definition(package)

        mock_log.exception.assert_called_once_with(
            'Unable to load package from cache. Clean-up.')
        mock_log.exception.reset_mock()

        # Test that the second instance of the exception is caught.
        mock_os.path.isdir.return_value = [False, True]

        self.loader._get_package_by_definition(package)

        mock_log.exception.assert_called_once_with(
            'Unable to load package from cache. Clean-up.')
        os.remove(temp_directory)

    @testtools.skipIf(os.name == 'nt', "Doesn't work on Windows")
    def test_get_package_by_definition_except_http_exception(self):
        fqn = 'io.murano.apps.test'
        path, _ = utils.compose_package(
            'test', self.location, archive_dir=self.location)

        package = mock.MagicMock()
        package.fully_qualified_name = fqn
        package.id = '123'
        package.version = '0.0.1'

        self.murano_client.packages.download.side_effect =\
            muranoclient_exc.HTTPException

        expected_error_msg = 'Error loading package id {0}:'.format(package.id)
        with self.assertRaisesRegex(pkg_exc.PackageLoadError,
                                    expected_error_msg):
            self.loader._get_package_by_definition(package)

    @testtools.skipIf(os.name == 'nt', "Doesn't work on Windows")
    @mock.patch('murano.engine.package_loader.LOG')
    @mock.patch('murano.engine.package_loader.tempfile')
    def test_get_package_by_definition_except_io_error(self, mock_tempfile,
                                                       mock_log):
        fqn = 'io.murano.apps.test'
        path, _ = utils.compose_package(
            'test', self.location, archive_dir=self.location)
        with open(path, 'rb') as f:
            package_data = f.read()

        package = mock.MagicMock()
        package.fully_qualified_name = fqn
        package.id = '123'
        package.version = '0.0.1'

        self.murano_client.packages.download = mock.MagicMock(
            return_value=package_data)
        mock_package_file = mock.MagicMock(
            write=mock.MagicMock(side_effect=IOError))
        mock_package_file.configure_mock(name='test_package_file')
        mock_tempfile.NamedTemporaryFile().__enter__.return_value =\
            mock_package_file

        expected_error_msg = 'Unable to extract package data for {0}'\
                             .format(package.id)
        with self.assertRaisesRegex(pkg_exc.PackageLoadError,
                                    expected_error_msg):
            self.loader._get_package_by_definition(package)

    def test_try_cleanup_cache_with_null_package_directory(self):
        # Test null package directory causes early return.
        result = self.loader.try_cleanup_cache(None, None)
        self.assertIsNone(result)

    @mock.patch('murano.engine.package_loader.shutil')
    @mock.patch('murano.engine.package_loader.m_utils')
    @mock.patch('murano.engine.package_loader.usage_mem_locks')
    @mock.patch('murano.engine.package_loader.LOG')
    @mock.patch('murano.engine.package_loader.os')
    def test_try_cleanup_cache_except_os_error(self, mock_os, mock_log,
                                               mock_usage_mem_locks, *args):
        # Test first instance of OSError is handled.
        mock_os.listdir.side_effect = OSError

        result = self.loader.try_cleanup_cache(None, None)
        self.assertIsNone(result)

        # Test second instance of OSError is handled.
        mock_os.reset_mock()
        mock_os.listdir.return_value = {'1', '2'}
        mock_os.listdir.side_effect = None
        mock_os.remove.side_effect = OSError
        mock_usage_mem_locks.__getitem__().write_lock().__enter__.\
            return_value = True

        result = self.loader.try_cleanup_cache('test_package_directory', None)
        self.assertIsNone(result)

        self.assertIn("Couldn't delete lock file:",
                      str(mock_log.warning.mock_calls[0]))


class TestCombinedPackageLoader(base.MuranoTestCase):

    def setUp(self):
        super(TestCombinedPackageLoader, self).setUp()

        location = os.path.dirname(__file__)
        self.override_config('load_packages_from', [location], 'engine')
        self.execution_session = mock.MagicMock()
        self.loader = package_loader.CombinedPackageLoader(
            self.execution_session)
        self._patch_api_loader()

        self.local_pkg_name = 'io.murano.test.MyTest'
        self.api_pkg_name = 'test.mpl.v1.app.Thing'

    def _patch_api_loader(self):
        self.api_loader_patcher = mock.patch.object(
            self.loader, 'api_loader', return_value=mock.MagicMock())
        self.api_loader_patcher.start()

    def _unpatch_api_loader(self):
        self.api_loader_patcher.stop()

    def test_loaders_initialized(self):
        self.assertEqual(1, len(self.loader.directory_loaders),
                         'One directory class loader should be initialized'
                         ' since there is one valid murano pl package in the'
                         ' provided directory in config.')
        self.assertIsInstance(self.loader.directory_loaders[0],
                              package_loader.DirectoryPackageLoader)

    def test_get_package_by_class_directory_loader(self):
        spec = semantic_version.Spec('*')
        result = self.loader.load_class_package(self.local_pkg_name, spec)
        self.assertIsInstance(result, dsl_package.MuranoPackage)

    def test_get_package_by_name_directory_loader(self):
        spec = semantic_version.Spec('*')
        result = self.loader.load_package(self.local_pkg_name, spec)
        self.assertIsInstance(result, dsl_package.MuranoPackage)

    def test_get_package_by_class_api_loader(self):
        spec = semantic_version.Spec('*')
        self.loader.load_package(self.api_pkg_name, spec)

        self.loader.api_loader.load_package.assert_called_with(
            self.api_pkg_name, spec)

    def test_get_package_api_loader(self):
        spec = semantic_version.Spec('*')
        self.loader.load_class_package(self.api_pkg_name, spec)

        self.loader.api_loader.load_class_package.assert_called_with(
            self.api_pkg_name, spec)

    def test_register_package(self):
        test_package = mock.MagicMock()
        self.loader.register_package(test_package)
        self.loader.api_loader.register_package.assert_called_once_with(
            test_package)

    def test_import_fixation_table(self):
        self.loader.directory_loaders = [
            mock.MagicMock(), mock.MagicMock(), mock.MagicMock()
        ]
        test_fixations = {
            'test_package_1': [semantic_version.Version('1.1.0')],
            'test_package_2': [semantic_version.Version('2.1.0'),
                               semantic_version.Version('2.4.3')]
        }
        self.loader.import_fixation_table(test_fixations)
        self.loader.api_loader.import_fixation_table.assert_called_once_with(
            test_fixations)
        for loader in self.loader.directory_loaders:
            loader.import_fixation_table.assert_called_once_with(
                test_fixations)

    def test_compact_fixation_table(self):
        self.loader.directory_loaders = [
            mock.MagicMock(), mock.MagicMock(), mock.MagicMock()
        ]
        self.loader.compact_fixation_table()
        self.loader.api_loader.compact_fixation_table.assert_called_once_with()
        for loader in self.loader.directory_loaders:
            loader.compact_fixation_table.assert_called_once_with()

    def test_export_fixation_table(self):
        self._unpatch_api_loader()

        test_fixations = {
            'test_package_1': [semantic_version.Version('1.1.1')],
            'test_package_2': [semantic_version.Version('2.2.2'),
                               semantic_version.Version('3.3.3')]
        }
        self.loader.api_loader.import_fixation_table(test_fixations)

        expected_table = {'test_package_1': ['1.1.1'],
                          'test_package_2': sorted(['2.2.2', '3.3.3'])}
        table = self.loader.export_fixation_table()
        table['test_package_2'] = sorted(table['test_package_2'])

        self.assertEqual(sorted(expected_table.items()),
                         sorted(table.items()))

        test_fixations = {
            'test_package_1': [semantic_version.Version('4.4.4')],
            'test_package_2': [semantic_version.Version('5.5.5'),
                               semantic_version.Version('6.6.6')],
            'test_package_3': [semantic_version.Version('7.7.7')]
        }
        self.loader.directory_loaders[0].import_fixation_table(test_fixations)

        expected_table = {'test_package_1': ['1.1.1', '4.4.4'],
                          'test_package_2': ['2.2.2', '3.3.3', '5.5.5',
                                             '6.6.6'],
                          'test_package_3': ['7.7.7']}
        table = self.loader.export_fixation_table()
        for key, value in table.items():
            table[key] = sorted(value)

        self.assertEqual(sorted(expected_table.items()),
                         sorted(table.items()))



# Copyright 2016 AT&T Corp
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from murano.db import models
from murano.db.services import stats as statistics
from murano.db import session as db_session
from murano.tests.unit import base


class TestStats(base.MuranoWithDBTestCase):

    def setUp(self):
        super(TestStats, self).setUp()

        self.stats_list = []
        self.stats_kwargs = {
            'host': 'test_host',
            'request_count': 5,
            'error_count': 2,
            'average_response_time': 0.5,
            'requests_per_tenant': 1,
            'cpu_count': 4,
            'cpu_percent': 0.3
        }
        self.stats_kwargs_2 = {
            'host': 'test_host_2',
            'request_count': 6,
            'error_count': 3,
            'average_response_time': 0.6,
            'requests_per_tenant': 2,
            'cpu_count': 5,
            'cpu_percent': 0.4
        }

    def tearDown(self):
        super(TestStats, self).tearDown()

        unit = db_session.get_session()
        for stat in self.stats_list:
            with unit.begin():
                unit.delete(stat)

    def _create_stats(self, which_kwargs):
        stats = models.ApiStats()
        for key, val in which_kwargs.items():
            setattr(stats, key, val)
        stats.requests_per_second = 1.2
        stats.errors_per_second = 2.3

        unit = db_session.get_session()
        with unit.begin():
            unit.add(stats)

        self.assertIsNotNone(stats)
        self.stats_list.append(stats)

        return stats

    def _are_stats_equal(self, x, y, check_type=False):
        comparison_attributes = ['host', 'request_count', 'error_count',
                                 'average_response_time',
                                 'requests_per_tenant',
                                 'requests_per_second',
                                 'errors_per_second',
                                 'cpu_count', 'cpu_percent']
        if check_type:
            if type(x) != type(y):
                return False
        for attr in comparison_attributes:
            try:
                if str(getattr(x, attr)) != str(getattr(y, attr)):
                    return False
            except AttributeError:
                if str(x[attr]) != str(getattr(y, attr)):
                    return False
        return True

    def test_create(self):
        statistics.Statistics.create(**self.stats_kwargs)

        unit = db_session.get_session()
        retrieved_stats = None
        with unit.begin():
            retrieved_stats = unit.query(models.ApiStats)\
                .order_by(models.ApiStats.id.desc()).first()

        self.stats_kwargs['requests_per_second'] = 0.0
        self.stats_kwargs['errors_per_second'] = 0.0
        self.assertIsNotNone(retrieved_stats)
        self.assertTrue(
            self._are_stats_equal(self.stats_kwargs, retrieved_stats))

    def test_update(self):
        """Note: this test expects to test update functionality.

        However, the current implementation of Statistics.update() does not
        actually update the host of the statistics object passed in.
        It just saves the object that is passed in, which appears to contradict
        its intended use case.
        """
        stats = models.ApiStats()
        for key, val in self.stats_kwargs.items():
            setattr(stats, key, val)
        statistics.Statistics.update('test_host', stats)

        unit = db_session.get_session()
        retrieved_stats = None
        with unit.begin():
            retrieved_stats = unit.query(models.ApiStats)\
                .order_by(models.ApiStats.id.desc()).first()
        self.assertIsNotNone(retrieved_stats)
        self.assertTrue(
            self._are_stats_equal(stats, retrieved_stats, check_type=True))

    def test_get_stats_by_id(self):
        stats = self._create_stats(self.stats_kwargs)
        retrieved_stats = statistics.Statistics.get_stats_by_id(stats.id)
        self.assertIsNotNone(retrieved_stats)
        self.assertTrue(
            self._are_stats_equal(stats, retrieved_stats, check_type=True))

    def test_get_stats_by_host(self):
        stats = self._create_stats(self.stats_kwargs)
        retrieved_stats = statistics.Statistics.get_stats_by_host(stats.host)
        self.assertIsNotNone(retrieved_stats)
        self.assertTrue(
            self._are_stats_equal(stats, retrieved_stats, check_type=True))

    def test_get_all(self):
        stats = self._create_stats(self.stats_kwargs)
        retrieved_stats = statistics.Statistics.get_all()
        self.assertIsInstance(retrieved_stats, list)
        self.assertEqual(1, len(retrieved_stats))
        self.assertTrue(
            self._are_stats_equal(stats, retrieved_stats[0], check_type=True))

        stats = self._create_stats(self.stats_kwargs_2)
        retrieved_stats = statistics.Statistics.get_all()
        self.assertIsInstance(retrieved_stats, list)
        self.assertEqual(2, len(retrieved_stats))
        self.assertTrue(
            self._are_stats_equal(stats, retrieved_stats[1], check_type=True))
